/* bdb-LU.pn.c */

/* This program is the node processor software for a parallel version
of the Shur complement LU factorization algorithm for
block-diagonal-bordered form sparse matrices.  */

/* special considerations for this implementation:

1) Matrices are in symmetric block-diagonal-bordered form

2) any number of processors can be specified up to an including the
total number of processors on the CM-5

3) it is assumed that ordering and load balancing are performed a
priori

4) the last block is considered sparse and solved with a sparse
version of kij-saxpy LU factorization


/* Naming convention used throughout this software

i - row number
j - column number
k - number along the diagonal

Lij   - any value in a strictly lower diagonal matrix block
Lkk   - value of the kth element on the lower diagonal
        stored in the location diag[k].Value
Lix   - row in a strictly lower diagonal matrix block
Lxk   - column in a strictly lower diagonal matrix block
        only accessed when updating a column
LBix  - partial row in a lower matrix border
LBxk  - column in a lower matrix border
        only accessed when updating a column

Uij   - any value in a strictly upper diagonal matrix block
Ukk   - value of the kth element on the upper diagonal
        always equal to 1.0, so it is not expliticly stored
Uxj   - column in a strictly upper diagonal matrix block
Ukx   - row in a strictly upper diagonal matrix block
        only accessed when updating a row
UBxj  - partial column in an upper matrix border
UBkx  - row in an upper matrix border
        only accessed when updating a row

Uij   - any value in a strictly upper diagonal matrix block
Ukk   - value of the kth element on the upper diagonal
        always equal to 1.0, so it is not expliticly stored
Uxj   - column in a strictly upper diagonal matrix block
Ukx   - row in a strictly upper diagonal matrix block
        only accessed when updating a row
UBxj  - partial column in an upper matrix border
UBkx  - row in an upper matrix border
        only accessed when updating a row

locLijUjk    - function to find the implicit storage locations
               of Lij and Ujk, where i and k are given, and
               (Lij != 0.) && (Ujk != 0.) && (j < k)
locLBijUjk   - function to find the implicit storage locations
               of LBij and Ujk, where i and k are given, and
               (LBij != 0.) && (Ujk != 0.) && (j < k)
locLkiUBij   - function to find the implicit storage locations
               of Lki and UBij, here k and j are given, and
               (Lki != 0.) && (UBij != 0.) && (j < k)
locLBijUBjk  - function to find the implicit storage locations
               of LBij and UBjk, where i and k are given, and
               (LBij != 0.) && (UBjk != 0.) && (j < k)
*/

#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <malloc.h>
#include "cmmd.h"
#include "interface.h"
#include "bdb_LU_struct.h"
#include "bdb_LU_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
#define EPSILON    0.0000001
#define CMAML_RPC  0            /* 1 - YES or 0 - NO */

#if 0
#define SP1
#endif
#define MPI

#define CMMD
/********************  For Active Message scopy  *********************/
#define false      0
#define true       1
#define rport      10
/*---------------------------------------------------------*/
typedef struct Node_Timer
{
   int             pn_id;
   double          elapsed;
   double          busy;
   double          idle;
}               node_timer;

node_timer      node_tim;

/*---------------------------------------------------------*/
Block          *blocks;         /* blocks[]    */

Diag           *diag;           /* diag[]      */

BlockRow       *blockrow;       /* blockrow[]  */

BlockCol       *blockcol;       /* blockcol[]  */

LowerBorder    *lborder;        /* lborder[]   */

UpperBorder    *uborder;        /* uborder[]   */

BorderRow      *borderrow;      /* borderrow[] */

BorderCol      *bordercol;      /* bordercol[] */

int            *lbbufsizeFac;

complex        *x;              /* x[]         */

complex        *y;              /* y[]         */

complex        *b;              /* b[]         */

complex        *lx;             /* lx[]        */

complex        *ly;             /* ly[]        */

complex        *lb;             /* lb[]        */

int             n;              /* matrix dimension       */

int             nblocks;        /* # of blocks            */

int             nblockrow;      /* dimension of blockrow  */

int             nblockcol;      /* dimension of blockcol  */

int             nlborder;       /* dimension of lborder   */

int             nuborder;       /* dimension of uborder   */

int             nborderrow;     /* dimension of borderrow */

int             nbordercol;     /* dimension of bordercol */

int             nlastblock;

int             lastblockstart;

int             lastblockend;

int             nprocs;

int             procnum;

int             ipollULBCMAML;  /* update last block */

int             ipollFRBCMAML;  /* forward reduce last block */

/*---------------------------------------------------------*/
BBuf            bbuf;

/*---------------------------------------------------------*/
/* variables for kij-saxpy code */
LBDiag         *lbdiag;

LBRow          *lbrowU;

LBRow          *lbrowL;

LBCol          *lbcolU;

LBCol          *lbcolL;

LBRow          *lbrowUrcv;

LBDiagColPt    *lbdiagcolpt;

int             nlb;            /* dimension of the last block */

int             nlbdiag;

int             nlbrowL;

int             nlbrowU;

int             nlbbufsizeFac;

int             nb;             /* block-width */

int             nbks;

int             bksthisproc;

int             maxbksperproc;

int             rowsthisproc;

int            *nupdatesULB;

int            *nupdatesFRB;

int             nflopsl, nflopsu, nflopsd, nflops;

LBRow         **lbupdate;

LBRow          *lbupdatetmp;

LBRow          *lbupdatercv;

int             lbupdatercv_size;

int            *nlbupdate;

LBUpdateFr    **lbupdatefr;

LBUpdateFr     *lbupdatefrtmp;

LBUpdateFr     *lbupdatefrrcv;

int             lbupdatefrrcv_size;

int            *nlbupdatefr;

/*---------------------------------------------------------*/
typedef struct DBLObj
{
   unsigned int    i1;
   unsigned int    i2;
}               DBL;

/*---------------------------------------------------------*/
/********************  For Active Message scopy  *********************/

int             rpc_ack = false;/* flag for CMAML_rpc */

int             scopy_ack = false;      /* flag for CMAML_scopy */

int             dest_chk;       /* for checking readiness-signal from
                                 * destination */

/********************************************************************/
/* function prototypes */

static void     InputData();

void            Factor();

void            FactorIndependentBlocks();

void            UpdateLkk();

void            locLijUjk();

void            UpdateLxk();

void            UpdateLBxk();

void            locLBijUjk();

void            UpdateUkx();

void            UpdateUBkx();

void            locLkiUBij();

void            UpdateLastBlockCMAML();

void            StoreHandlerULBCMAML();

void            locLBijUBjk();

void            UpdateLastBlock();

void            SendHandlerCMMD();

void            ReceiveHandlerCMMD();

void            FactorLastBlock();

void            BlockSparseFac();

void            BlockSparsekijFac();

void            SparsekijLUSeq();

void            SparsekijLUPar();

void            SparsekijLUUpdate();

void            activesendSLB();

void            activerecvSLB();

void            rpc_handler();

void            scopy_handler();

void            ForwardReduction();

void            ForwardReduceBlocks();

void            ForwardReduceBorderCMAML();

void            StoreHandlerFRBCMAML();

void            ForwardReduceBorder();

void            ForwardReduceLastBlock();

void            BlockSparseFr();

void            SparseFrSeq();

void            SparseFrPar();

void            SparseFrUpdate();

void            BackwardSubstitution();

void            BackSubLastBlock();

void            BlockSparseBs();

void            SparseBsSeq();

void            SparseBsPar();

void            SparseBsUpdate();

void            BackSubBlocks();

node_main(argc, argv)
int             argc;

char          **argv;

{
   CMMD_enable_host();
   enable_diag_write();
   procnum = CMMD_self_address();

#if CMAML_RPC
/*-----  Initialize rport for CMAML_scopy  -----*/
   CMAML_disable_interrupts();
   if (CMAML_allocate_this_rport(rport) <= 0)
      CMMD_error("ERROR: Failed to allocate receive rport %d\n", rport);
   CMAML_set_rport_handler(rport, scopy_handler);
   CMAML_set_rport_handler_arg2(rport, 0);
   CMAML_enable_interrupts();
/*---------------------------------------------*/
#endif

   printf("InputData() \n");
   fflush(stdout);
   InputData();
   rpc_ack = false;             /* clear flag for CMAML_rpc */
   scopy_ack = false;           /* clear flag for CMAML_scopy */

   printf("Factor \n");
   fflush(stdout);
   Factor();
   rpc_ack = false;             /* clear flag for CMAML_rpc */
   scopy_ack = false;           /* clear flag for CMAML_scopy */

   printf("ForwardReduction \n");
   fflush(stdout);
   ForwardReduction();
   rpc_ack = false;             /* clear flag for CMAML_rpc */
   scopy_ack = false;           /* clear flag for CMAML_scopy */

   printf("BackwardSubstitution \n");
   fflush(stdout);
   BackwardSubstitution();
   rpc_ack = false;             /* clear flag for CMAML_rpc */
   scopy_ack = false;           /* clear flag for CMAML_scopy */
   CMMD_sync_with_host();
}                               /* end main() */



/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
static void     InputData()
{
   int             i, j, k, l, m, bufsize, bytesrec;
   complex         e1, e2;

/* (I) set up variables at this processor */
   CMMD_receive_bc_from_host(&bbuf, sizeof(BBuf));
   nlb = bbuf.nlb;
   nb = bbuf.nb;
   nprocs = bbuf.nprocs;
   printf("nlb=%d - nb=%d - nprocs=%d - procnum=%d  \n", nlb, nb, nprocs, procnum);

   if (procnum < nprocs)
   {
      nbks = ((nlb + nb - 1) / nb);
      bksthisproc = (nbks / nprocs);
      if (procnum < (nbks % nprocs))
         bksthisproc++;
      maxbksperproc = (nbks + nprocs - 1) / nprocs;
      rowsthisproc = (bksthisproc * nb);
      if ((procnum == (((nbks - 1) % nprocs))) && (nlb < (nbks * nb)))
         rowsthisproc -= ((nbks * nb) - nlb);

      printf("nbks=%d - bksthisproc=%d - maxbksperproc=%d - rowsthisproc=%d   \n",
         nbks, bksthisproc, maxbksperproc, rowsthisproc);
      fflush(stdout);

/* (I-1) receive the block identifer array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPBLOCKS, &nblocks, sizeof(int));
/*
printf("receive the block identifer array of structures --- nblocks=%d \n",nblocks);
fflush(stdout);
*/
      blocks = (Block *) calloc((nblocks + 1), sizeof(Block));
      if (blocks == NULL)
      {
         CMMD_error("can't allocate memory for blocks[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PBLOCKS, blocks, ((nblocks + 1) * sizeof(Block)));
/*
if (procnum == 0) {
  for (i=1;i<=nblocks;i++) {
    printf("%d \t%d %d %d %d %d %d \n",i,
               blocks[i].BlockID,
               blocks[i].BlockSz,
               blocks[i].ProcNum,
               blocks[i].DiagPtr,
               blocks[i].LBPtr,
               blocks[i].UBPtr);
    fflush(stdout); }}
*/
/* (I-2) receive the matrix diagonal array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NP, &n, sizeof(int));
/*
printf("receive the matrix diagonal array of structures --- n=%d \n",n);
fflush(stdout);
*/
      diag = (Diag *) calloc((n + 1), sizeof(Diag));
      if (diag == NULL)
      {
         CMMD_error("can't allocate memory for diag[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PDIAG, diag, ((n + 1) * sizeof(Diag)));
/*
if (procnum == 0) {
  for (i=1;i<=n;i++) {
    printf("%d \t(%f,%fi) %d %d %d %d %d %d \n",i,
             diag[i].Value.real,
             diag[i].Value.imag,
             diag[i].PtrLix,
             diag[i].NextLxk,
             diag[i].NextLBxk,
             diag[i].PtrUxj,
             diag[i].NextUkx,
             diag[i].NextUBkx);
    fflush(stdout); }}
*/
/* (I-3) receive the block row storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPBLOCKROW, &nblockrow, sizeof(int));
/*
printf("receive the block row storage array of structures --- nblockrow=%d \n",nblockrow);
fflush(stdout);
*/
      blockrow = (BlockRow *) calloc((nblockrow + 1), sizeof(BlockRow));
      if (blockrow == NULL)
      {
         CMMD_error("can't allocate memory for blockrow[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PBLOCKROW, blockrow, ((nblockrow + 1) * sizeof(BlockRow)));
/*
if (procnum == 0) {
  for (i=1;i<=nblockrow;i++) {
    printf("%d \t(%f,%f) %d %d %d \n",i,
                    blockrow[i].Value.real,
                    blockrow[i].Value.imag,
                    blockrow[i].i,
                    blockrow[i].j,
                    blockrow[i].NextLxk);
    fflush(stdout); } }
*/
/* (I-4) receive the block column storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPBLOCKCOL, &nblockcol, sizeof(int));
/*
printf("receive the block Column storage array of structures --- nblockcol=%d \n",nblockcol);
fflush(stdout);
*/
      blockcol = (BlockCol *) calloc((nblockcol + 1), sizeof(BlockCol));
      if (blockcol == NULL)
      {
         ("can't allocate memory for blockcol[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PBLOCKCOL, blockcol, ((nblockcol + 1) * sizeof(BlockCol)));
/*
if (procnum == 1) {
  for (i=1;i<=nblockcol;i++) {
    printf("%d \t(%f,%f) %d %d %d \n",i,
                    blockcol[i].Value.real,
                    blockcol[i].Value.imag,
                    blockcol[i].i,
                    blockcol[i].j,
                    blockcol[i].NextUkx);
    fflush(stdout); }}
*/
/* (I-5) receive the lower border row storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPLBORDER, &nlborder, sizeof(int));
/*
printf("receive the lower border row storage array of structures --- nlborder=%d \n",nlborder);
fflush(stdout);
*/
      lborder = (LowerBorder *) calloc((nlborder + 1), sizeof(LowerBorder));
      if (lborder == NULL)
      {
         CMMD_error("can't allocate memory for lborder[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PLBORDER, lborder, ((nlborder + 1) * sizeof(LowerBorder)));
/*
if (procnum == 10) {
  for (i=1;i<=nlborder;i++) {
    printf("%d \t%d %d \n",i,
                  lborder[i].i,
                  lborder[i].PtrLBix);
    fflush(stdout); }}
*/
/* (I-6) receive the upper border column storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPUBORDER, &nuborder, sizeof(int));
/*
printf("receive the upper border column storage array of structures --- nuborder=%d \n",nuborder);
fflush(stdout);
*/
      uborder = (UpperBorder *) calloc((nuborder + 1), sizeof(UpperBorder));
      if (uborder == NULL)
      {
         CMMD_error("can't allocate memory for uborder[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PUBORDER, uborder, ((nuborder + 1) * sizeof(UpperBorder)));
/*
if (procnum == 10) {
  for (i=1;i<=nlborder;i++) {
    printf("%d \t%d %d \n",i,
                  uborder[i].j,
                  uborder[i].PtrUBxj);
    fflush(stdout); }}
*/
/* (I-7) receive the border row storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPBORDERROW, &nborderrow, sizeof(int));
/*
printf("receive the border row storage array of structures --- nborderrow=%d \n",nborderrow);
fflush(stdout);
*/
      borderrow = (BorderRow *) calloc((nborderrow + 1), sizeof(BorderRow));
      if (borderrow == NULL)
      {
         CMMD_error("can't allocate memory for borderrow[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PBORDERROW, borderrow, ((nborderrow + 1) * sizeof(BorderRow)));
/*
if (procnum == 1) {
  for (i=1;i<=nborderrow;i++) {
    printf("%d \t(%f,%f) %d %d %d \n",i,
            borderrow[i].Value.real,
            borderrow[i].Value.imag,
            borderrow[i].i,
            borderrow[i].j,
            borderrow[i].NextLBxk);
    fflush(stdout); }}
*/
/* (I-8) receive the border column storage array of structures */
      CMMD_receive(CMMD_host_node(), RECEIVE_NPBORDERCOL, &nbordercol, sizeof(int));
/*
printf("receive the border column storage array of structures --- nbordercol=%d \n",nbordercol);
fflush(stdout);
*/
      bordercol = (BorderCol *) calloc((nbordercol + 1), sizeof(BorderCol));
      if (bordercol == NULL)
      {
         CMMD_error("can't allocate memory for bordercol[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PBORDERCOL, bordercol, ((nbordercol + 1) * sizeof(BorderCol)));
/*
if (procnum == 1) {
  for (i=1;i<=nbordercol;i++) {
    printf("%d \t(%f,%f) %d %d %d \n",i,
            bordercol[i].Value.real,
            bordercol[i].Value.imag,
            bordercol[i].i,
            bordercol[i].j,
            bordercol[i].NextUBkx);
    fflush(stdout); }}
*/

/* (II-1) receive this processor's portion of the last diagonal block */
/*
printf("(II) receive this processor's portion of the last diagonal block - diagonal\n");
printf("\tnlb=%d - bksthisproc=%d - nb=%d\n",nlb,bksthisproc,nb);
fflush(stdout);
*/
      CMMD_receive(CMMD_host_node(), RECEIVE_NLBDIAG, &nlbdiag, sizeof(int));
      CMMD_receive(CMMD_host_node(), RECEIVE_NB, &nb, sizeof(int));
/*
printf("receive the last diagonal block - diagonal --- nlbdiag=%d \n",nlbdiag);
fflush(stdout);
*/
      lbdiag = (LBDiag *) calloc((nlbdiag + 1), sizeof(LBDiag));
      if (lbdiag == NULL)
      {
         CMMD_error("can't allocate memory for lbdiag[] \n");
      }                         /* end if */

      bufsize = ((nlbdiag + 1) * sizeof(LBDiag));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBDIAG, lbdiag, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlbdiag;i++)
    printf("%d \t(%f,%f) %d %d \n",i,
             lbdiag[i].Value.real,
             lbdiag[i].Value.imag,
             lbdiag[i].PtrUix,
             lbdiag[i].PtrUxj,
             lbdiag[i].PtrLix,
             lbdiag[i].PtrLxj,
             lbdiag[i].k);
    fflush(stdout); }
*/
/* (II-2) receive this processor's portion of the last diagonal block - rows in the lower diagonal */
/*
printf("(II) receive this processor's portion of the last diagonal block - rows in the lower diagonal \n");
fflush(stdout);
*/
      CMMD_receive(CMMD_host_node(), RECEIVE_NLBROWL, &nlbrowL, sizeof(int));
/*
printf("receive the last diagonal block - rows in the lower diagonal --- nlbrowL=%d \n",nlbrowL);
fflush(stdout);
*/
      lbrowL = (LBRow *) calloc((nlbrowL + 1), sizeof(LBRow));
      if (lbrowL == NULL)
      {
         CMMD_error("can't allocate memory for lbrowL[] \n");
      }                         /* end if */
      bufsize = ((nlbrowL + 1) * sizeof(LBRow));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBROWL, lbrowL, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlbrowU;i++) {
    printf("%d \t(%f,%f) %d %d \n",i,
             lbrowL[i].Value.real,
             lbrowL[i].Value.imag,
             lbrowL[i].i,
             lbrowL[i].j);
    fflush(stdout); } }


/*
   (II-3) receive this processor's portion of the last diagonal block - rows in the upper diagonal
*/
/*
printf("(II) receive this processor's portion of the last diagonal block - rows in the upper diagonal \n");
fflush(stdout);
*/
      CMMD_receive(CMMD_host_node(), RECEIVE_NLBROWU, &nlbrowU, sizeof(int));
/*
printf("receive the last diagonal block - rows in the upper diagonal --- nlbrowU=%d \n",nlbrowU);
fflush(stdout);
*/
      lbrowU = (LBRow *) calloc((nlbrowU + 1), sizeof(LBRow));
      if (lbrowU == NULL)
      {
         CMMD_error("can't allocate memory for lbrowU[] \n");
      }                         /* end if */
      bufsize = ((nlbrowU + 1) * sizeof(LBRow));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBROWU, lbrowU, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlbrowU;i++) {
    printf("%d \t(%f,%f) %d %d \n",i,
             lbrowU[i].Value.real,
             lbrowU[i].Value.imag,
             lbrowU[i].i,
             lbrowU[i].j);
    fflush(stdout); } }


/*
   (II-4) receive this processor's portion of the column pointers in the lower diagonal
*/
/*
printf("(II) receive this processor's portion of the column pointers in the lower diagonal \n");
fflush(stdout);
*/
      lbcolL = (LBCol *) calloc((nlbrowL + 1), sizeof(LBCol));
      if (lbcolL == NULL)
      {
         CMMD_error("can't allocate memory for lbcolL[] \n");
      }                         /* end if */
      bufsize = ((nlbrowL + 1) * sizeof(LBCol));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBCOLL, lbcolL, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlbrowL;i++) {
    printf("%d \t%d \n",i,
             lbcolL[i].PtrNext);
    fflush(stdout); } }


/*
   (II-5) receive this processor's portion of the column pointers in the lower diagonal
*/
/*
printf("(II) receive this processor's portion of the column pointers in the lower diagonal \n");
fflush(stdout);
*/
      lbcolU = (LBCol *) calloc((nlbrowU + 1), sizeof(LBCol));
      if (lbcolU == NULL)
      {
         CMMD_error("can't allocate memory for lbcolU[] \n");
      }                         /* end if */
      bufsize = ((nlbrowU + 1) * sizeof(LBCol));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBCOLU, lbcolU, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlbrowU;i++) {
    printf("%d \t%d \n",i,
             lbcolU[i].PtrNext);
    fflush(stdout); } }


/*
   (II-6) receive this processor's structure of column pointers for all columns for FR
*/
/*
printf("(II) receive this processor's structure of column pointers for all columns for FR \n");
fflush(stdout);
*/
      lbdiagcolpt = (LBDiagColPt *) calloc((nlb + 1), sizeof(LBDiagColPt));
      if (lbdiagcolpt == NULL)
      {
         CMMD_error("can't allocate memory for lbdiagcolpt[] \n");
      }                         /* end if */
      bufsize = ((nlb + 1) * sizeof(LBDiagColPt));
      CMMD_receive(CMMD_host_node(), RECEIVE_LBDIAGCOLPT, lbdiagcolpt, bufsize);
/*
if (procnum == 0) {
  for (i=0;i<=nlb;i++) {
    printf("%d \t%d %d %d \n",i,
             lbdiagcolpt[i].k,
             lbdiagcolpt[i].PtrUxj,
             lbdiagcolpt[i].PtrLxj);
    fflush(stdout); } }


/*
   (II-7) the buffer sizes when receiving the broadcasts of data when factoring the last block
*/
      CMMD_receive(CMMD_host_node(), RECEIVE_NLBBUFSIZEFAC, &nlbbufsizeFac, sizeof(int));
/*
printf("receive the last diagonal block - lbbufsizeFac --- nlbrowU=%d \n",nlbrowU);
fflush(stdout);
*/
      lbbufsizeFac = (int *) calloc((nlbbufsizeFac + 1), sizeof(int));
      if (lbbufsizeFac == NULL)
      {
         CMMD_error("can't allocate memory for lbbufsizeFac[] \n");
      }                         /* end if */
      bufsize = (nlbbufsizeFac + 1) * sizeof(int);
      CMMD_receive(CMMD_host_node(), RECEIVE_LBBUFSIZEFAC, lbbufsizeFac, bufsize);

#if 0
#define DEBUG_InputData
#endif

#ifdef DEBUG_InputData
      for (i = 0; i <= nlbbufsizeFac; i++)
      {
         printf("%d receive InputData lbbufsizeFac %d %d\n",
            CMMD_self_address(), i, lbbufsizeFac[i]);
      }
      fflush(stdout);
#endif

/* (III-1) receive the b vector for the blocked portion of the matrix */
      printf("(III-1) receive a portion of the b vector \n");
      fflush(stdout);
      b = (complex *) calloc((n + 1), sizeof(complex));
      if (b == NULL)
      {
         CMMD_error("can't allocate memory for b[] \n");
      }                         /* end if */
      CMMD_receive(CMMD_host_node(), RECEIVE_PB, b, ((n + 1) * sizeof(complex)));

/* (III-2) receive a portion of the lb vector */

#ifdef DEBUG_InputData
      printf("(III-2) receive a portion of the lb vector - bksthisproc=%d \n",
         bksthisproc);
      fflush(stdout);
#endif

      lb = (complex *) calloc((nlb + 1), sizeof(complex));
      if (lb == NULL)
      {
         CMMD_error("can't allocate memory for lb[] \n");
      }                         /* end if */

      for (j = 0; j < bksthisproc; j++)
      {
         bufsize = (nb * sizeof(complex));

#ifdef DEBUG_InputData
         printf("(III-2) receive a portion of the lb vector - j=%d \n", j);
         fflush(stdout);
#endif

         CMMD_receive(CMMD_host_node(), RECEIVE_LB, &(lb[((j * nb * nprocs) + (nb * procnum) + 1)]), bufsize);

#ifdef DEBUG_InputData
         printf("(III-2) received a portion of the lb vector - j=%d \n", j);
         fflush(stdout);
#endif
      }                         /* end for j */

/* (IV-1) receive the row/col locations of the first and last locations in the
   last block */
      CMMD_receive(CMMD_host_node(), RECEIVE_LASTBKST, &lastblockstart, sizeof(int));
      CMMD_receive(CMMD_host_node(), RECEIVE_LASTBKND, &lastblockend, sizeof(int));
      printf("lastblockend=%d - lastblockstart=%d \n", lastblockend, lastblockstart);
      fflush(stdout);
      nlastblock = lastblockend - lastblockstart + 1;

/* (IV-2) the number of updates for factoring the last block for each proc */
      nupdatesULB = (int *) calloc((nprocs + 1), sizeof(int));
      if (nupdatesULB == NULL)
      {
         CMMD_error("can't allocate memory for nupdatesULB[] \n");
      }                         /* end if */
      bufsize = (nprocs + 1) * sizeof(int);

#ifdef DEBUG_InputData
      printf("%d: (IV-2) before CMMD_receive\n", CMMD_self_address());
#endif

      CMMD_receive(CMMD_host_node(), RECEIVE_NUPDATEULB, nupdatesULB, bufsize);

#ifdef DEBUG_InputData
      printf("%d: (IV-2) after CMMD_receive\n", CMMD_self_address());
#endif

      if (procnum == 1)
      {
         for (i = 0; i <= nprocs; i++)
         {
            printf("iproc=%d - nupdatesULB[%d]=%d \n", i, i, nupdatesULB[i]);
            fflush(stdout);
         }
      }

/* (IV-3) the number of updates for forward reduction of the last block for each proc */
      nupdatesFRB = (int *) calloc((nprocs + 1), sizeof(int));
      if (nupdatesFRB == NULL)
      {
         CMMD_error("can't allocate memory for nupdatesFRB[] \n");
      }                         /* end if */
      bufsize = (nprocs + 1) * sizeof(int);
      CMMD_receive(CMMD_host_node(), RECEIVE_NUPDATEFRB, nupdatesFRB, bufsize);

      if (procnum == 1)
      {
         for (i = 0; i <= nprocs; i++)
         {
            printf("iproc=%d - nupdatesFRB[%d]=%d \n", i, i, nupdatesFRB[i]);
            fflush(stdout);
         }
      }

/* (V) allocate space for the remaining vectors - sparse matrix software */
/*
printf("allocate space for the remaining vectors - sparse matrix software \n");
fflush(stdout);
*/
      y = (complex *) calloc((n + 1), sizeof(complex));
      if (y == NULL)
      {
         CMMD_error("can't allocate memory for y[] \n");
      }                         /* end if */

      x = (complex *) calloc((n + 1), sizeof(complex));
      if (x == NULL)
      {
         CMMD_error("can't allocate memory for x[] \n");
      }                         /* end if */

      lx = (complex *) calloc((nlb + 1), sizeof(complex));
      if (lx == NULL)
      {
         CMMD_error("can't allocate memory for lx[] \n");
      }                         /* end if */

      ly = (complex *) calloc((nlb + 1), sizeof(complex));
      if (ly == NULL)
      {
         CMMD_error("can't allocate memory for ly[] \n");
      }                         /* end if */

   }                            /* end if this is an active processor */
   printf("end InputData()\n");
}                               /* end InputData() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            Factor()
{
   int             i, iproc, bufsize, size;


#if 0
#define DEBUG_Factor
#endif

#ifdef DEBUG_Factor
   for (i = 1; i <= nlbrowL; i++)
   {
      printf("%d Factor lbrowL %d [%5d, %5d] = (%9.6f, %9.6f)\n",
         CMMD_self_address(), i, lbrowL[i].i, lbrowL[i].j,
         lbrowL[i].Value.real, lbrowL[i].Value.imag);
      printf("%d Factor lbrowU %d [%5d, %5d] = (%9.6f, %9.6f)\n",
         CMMD_self_address(), i, lbrowU[i].i, lbrowU[i].j,
         lbrowU[i].Value.real, lbrowU[i].Value.imag);
   }
   fflush(stdout);
#undef DEBUG_Factor
#endif

   CMMD_sync_with_host();
   if (procnum < nprocs)
   {
      printf("FactorIndependentBlocks \n");
      fflush(stdout);
      nflopsl = nflopsu = nflopsd = nflops = 0;

      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      FactorIndependentBlocks();

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);

#ifdef DEBUG_Factor
      printf("FactorIndependentBlocks: node = %d, elapsed = %g, busy = %g, \
idle = %g\n", node_tim.pn_id, node_tim.elapsed, node_tim.busy, node_tim.idle);
#endif

      CMMD_send(CMMD_host_node(), SEND_TIME_FIB, &node_tim, sizeof(node_timer));
   }                            /* end if an active processor */

#if 0
#define DEBUG_Factor
#endif

#ifdef DEBUG_Factor
   for (i = 1; i <= nlbrowL; i++)
   {
      printf("%d Factor lbrowL %d [%5d, %5d] = (%9.6f, %9.6f)\n",
         CMMD_self_address(), i, lbrowL[i].i, lbrowL[i].j,
         lbrowL[i].Value.real, lbrowL[i].Value.imag);
      printf("%d Factor lbrowU %d [%5d, %5d] = (%9.6f, %9.6f)\n",
         CMMD_self_address(), i, lbrowU[i].i, lbrowU[i].j,
         lbrowU[i].Value.real, lbrowU[i].Value.imag);
   }
   fflush(stdout);
#endif

#undef DEBUG_Factor
   CMMD_sync_with_host();

#if CMAML_RPC
   CMAML_disable_interrupts();
#endif

   if (procnum < nprocs)
   {
      printf("UpdateLastBlock() \n");
      fflush(stdout);
      if (CMAML_RPC == YES)
      {
         CMMD_node_timer_clear(0);
         CMMD_node_timer_start(0);

#if CMAML_RPC
         UpdateLastBlockCMAML();
#endif

         CMMD_node_timer_stop(0);
      }                         /* end if CMAML_RPC == YES */
      else
      {
         nlbupdate = (int *) calloc((nprocs + 1), sizeof(int));
         if (nlbupdate == NULL)
         {
            CMMD_error("can't allocate memory for nlbupdate[] \n");
         }                      /* end if */
         for (iproc = 0; iproc < nprocs; iproc++)
            nlbupdate[iproc] = 0;

         if (nprocs == 1)
            size = 1;
         else
            size = (nlborder * ((nlborder / nblocks) + 105));

         printf("size of transmission vectors = %d \n", size);
         fflush(stdout);

         lbupdatetmp = (LBRow *) calloc((size * (nprocs + 1)), sizeof(LBRow));
         if (lbupdatetmp == NULL)
         {
            CMMD_error("can't allocate memory for lbupdatetmp \n");
         }                      /* end if */

         lbupdate = (LBRow **) calloc((nprocs + 1), sizeof(LBRow *));
         if (lbupdate == NULL)
         {
            CMMD_error("can't allocate memory for lbupdate[][] \n");
         }                      /* end if */
         for (i = 0; i < (nprocs + 1); i++)
            lbupdate[i] = lbupdatetmp + i * size;

         lbupdatercv = (LBRow *) calloc(size, sizeof(LBRow));
         lbupdatercv_size = size * sizeof(LBRow);
         if (lbupdatercv == NULL)
         {
            CMMD_error("can't allocate memory for lbupdatercv[] \n");
         }                      /* end if */

         CMMD_node_timer_clear(0);
         CMMD_node_timer_start(0);

         UpdateLastBlock();

         CMMD_node_timer_stop(0);
         free(nlbupdate);
         free(lbupdatetmp);
         free(lbupdate);
         free(lbupdatercv);
      }                         /* end if CMAML_RP == NO */

      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_ULB, &node_tim, sizeof(node_timer));
   }                            /* end if an active processor */

#if CMAML_RPC
   CMAML_enable_interrupts();
#endif

#if 0
#define DEBUG_Factor
#endif

#ifdef DEBUG_Factor
   for (i = 1; i <= n; i++)
   {
      printf("%d Before Factor diag %d (%9.6f, %9.6f)\n",
         CMMD_self_address(), i, diag[i].Value.real, diag[i].Value.imag);
   }

   for (i = 1; i <= nlb; i++)
   {
      printf("%d Before Factor lbdiag %d (%9.6f, %9.6f)\n",
         CMMD_self_address(), i,
         lbdiag[i].Value.real, lbdiag[i].Value.imag);
   }
#endif

#undef DEBUG_Factor

   CMMD_sync_with_host();

   if (procnum < nprocs)
   {
      printf("FactorLastBlock \n");
      fflush(stdout);
      lbrowUrcv = (LBRow *) calloc((nprocs * nlb), sizeof(LBRow));
      if (lbrowUrcv == NULL)
      {
         CMMD_error("can't allocate memory for lbrowUrcv[] \n");
      }                         /* end if */
/*     nflopsl = nflopsu = nflopsd = 0; */

      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      FactorLastBlock();

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_FLB, &node_tim, sizeof(node_timer));
      free(lbrowUrcv);
/*
printf("FAC --- NUMBER OF FLOATING POINT OPS FOR THE LAST BLOCK ON THIS PROCESSOR --- nflopsu=%d - nflopsl=%d - nflopsd=%d \n",
               nflopsu,nflopsl,nflopsd);
*/

      printf("FAC --- NUMBER OF FLOATING POINT OPS FOR THIS PROCESSOR --- nflopsu=%d - nflopsl=%d - nflopsd=%d - nflops=%d \n",
         nflopsu, nflopsl, nflopsd, nflops);
      fflush(stdout);

#ifdef DEBUG_Factor
      for (i = 1; i <= nblockrow; i++)
      {
         printf("%d Factor blockrow %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, blockrow[i].i, blockrow[i].j,
            blockrow[i].Value.real, blockrow[i].Value.imag);
         printf("%d Factor blockcol %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, blockcol[i].i, blockcol[i].j,
            blockcol[i].Value.real, blockcol[i].Value.imag);
      }
      fflush(stdout);
#endif

#ifdef DEBUG_Factor
      for (i = 1; i <= nborderrow; i++)
      {
         printf("%d Factor borderrow %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, borderrow[i].i, borderrow[i].j,
            borderrow[i].Value.real, borderrow[i].Value.imag);
         printf("%d Factor bordercol %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, bordercol[i].i, bordercol[i].j,
            bordercol[i].Value.real, bordercol[i].Value.imag);
      }
      fflush(stdout);
#endif

#if 0
#define DEBUG_Factor
#endif

#ifdef DEBUG_Factor
      for (i = 1; i <= nlbrowL; i++)
      {
         printf("%d Factor lbrowL %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, lbrowL[i].i, lbrowL[i].j,
            lbrowL[i].Value.real, lbrowL[i].Value.imag);
         printf("%d Factor lbrowU %d [%5d, %5d] = (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, lbrowU[i].i, lbrowU[i].j,
            lbrowU[i].Value.real, lbrowU[i].Value.imag);
      }
      fflush(stdout);
#endif

#if 0
#define DEBUG_Factor
#endif

#ifdef DEBUG_Factor
      for (i = 1; i <= n; i++)
      {
         printf("%d After Factor diag %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, diag[i].Value.real, diag[i].Value.imag);
      }

      for (i = 1; i <= nlb; i++)
      {
         printf("%d After Factor lbdiag %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i,
            lbdiag[i].Value.real, lbdiag[i].Value.imag);
      }
#undef DEBUG_Factor
#endif
   }                            /* end if an active processor */
}                               /* end Factor() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/

void            FactorIndependentBlocks()
{
   int             i, k, l;
   complex         Lkk;

/* loop over the (nblocks) independent diagonal blocks
   the borders will be factorized after each diagonal block
   division by the value of Lkk occurs after a value is updated */

   for (l = 1; l <= nblocks; l++)
   {
/* loop along the diagonal in a block */
      for (i = (blocks[l - 1].DiagPtr + 1); i <= blocks[l].DiagPtr; i++)
      {
         k = i - blocks[0].DiagPtr;

#ifdef DEBUG_FactorIndependentBlocks
         printf("FactorIndependentBlocks() --- l=%d - k=%d \n", l, k);
         fflush(stdout);
#endif

/* (1) update the column Lxk */
         /* update the value on the diagonal */

#ifdef DEBUG_FactorIndependentBlocks
         printf("\tUpdateLkk(%d)\n", k);
         fflush(stdout);
#endif

         UpdateLkk(k);

         /*
          * for nonzero elements in the kth column of the strictly lower
          * diagonal porition of this block
          */

#ifdef DEBUG_FactorIndependentBlocks
         printf("\tUpdateLxk(%d) \n", k);
         fflush(stdout);
#endif

         UpdateLxk(k);

         /*
          * for nonzero elements in the kth column of the lower border of
          * this block
          */

#ifdef DEBUG_FactorIndependentBlocks
         printf("\tUpdateLBxk(%d,%d) \n", k, l);
         fflush(stdout);
#endif

         UpdateLBxk(k, l);

/* (2) update the row Ukx */

         /*
          * for nonzero elements in the kth row of the strictly upper
          * diagonal porition of this block
          */

#ifdef DEBUG_FactorIndependentBlocks
         printf("\tUpdateUkx(%d) \n", k);
         fflush(stdout);
#endif

         UpdateUkx(k);

         /*
          * for nonzero elements in the kth row of the upper border of this
          * block
          */

#ifdef DEBUG_FactorIndependentBlocks
         printf("\tUpdateUBkx(%d) \n", k);
         fflush(stdout);
#endif

         UpdateUBkx(k, l);

      }                         /* end for i - loop along diagonal in a block */
   }                            /* end for l - loop over independent blocks */
}                               /* end FactorIndependentBlocks() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateLkk(k)
int             k;

{
   int             locLij, locUjk;
   int             endLix, endUxk;
   complex         c1;

/* update the value on the diagonal (i=k) */
/* initialize the pointers into the data structures */
   locLij = (diag[(k - 1)].PtrLix + 1);
   locUjk = (diag[(k - 1)].PtrUxj + 1);
   endLix = diag[k].PtrLix;
   endUxk = diag[k].PtrUxj;
/* find the first pair of nonzero Lij and Ujk */
   locLijUjk(k, &locLij, &locUjk, endLix, endUxk);
   while ((locLij != 0) && (locUjk != 0))
   {
      /* diag[k].Value -= blockrow[locLij].Value * blockcol[locUjk].Value; */
      multiply(blockrow[locLij].Value, blockcol[locUjk].Value, c1);
      subtract(diag[k].Value, c1, diag[k].Value);

      locLij++;
      locUjk++;
      nflopsd++;
/* find the next pair of nonzero Lij and Ujk */
      locLijUjk(k, &locLij, &locUjk, endLix, endUxk);
   }                            /* end while (locLij and locUjk != 0) */
}                               /* end UpdateLkk() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            locLijUjk(k, locLij, locUjk, endLix, endUxk)
int             k, *locLij, *locUjk, endLix, endUxk;

{
   int             Search;
   int             lj, ui;

   if ((*locLij <= endLix) && (*locUjk <= endUxk))
   {
      lj = blockrow[*locLij].j - blocks[0].DiagPtr;
      ui = blockcol[*locUjk].i - blocks[0].DiagPtr;
      Search = YES;
/*
printf("\t\t\t\t1 - lj=%d - ui=%d - locLij=%d - locUjk=%d \n",lj,ui,*locLij,*locUjk);
fflush(stdout);
*/
      while (Search == YES)
      {
         if (lj == ui)
         {
            Search = NO;
         }                      /* end if lj == ui */
         else if (lj < ui)
         {
            while ((lj < ui) && (lj < k) && (*locLij <= endLix))
               lj = blockrow[++(*locLij)].j - blocks[0].DiagPtr;
/*
printf("\t\t\t\t2 - lj=%d - ui=%d - locLij=%d - locUjk=%d \n",lj,ui,*locLij,*locUjk);
fflush(stdout);
*/
            if ((lj >= k) || (*locLij > endLix))
            {
               Search = NO;
               *locLij = *locUjk = 0;
            }                   /* end if */
         }                      /* end if lj < ui */
         else if (lj > ui)
         {
            while ((lj > ui) && (ui < k) && (*locUjk <= endUxk))
               ui = blockcol[++(*locUjk)].i - blocks[0].DiagPtr;
/*
printf("\t\t\t\t3 - lj=%d - ui=%d - locLij=%d - locUjk=%d \n",lj,ui,*locLij,*locUjk);
fflush(stdout);
*/
            if ((ui >= k) || (*locUjk > endUxk))
            {
               Search = NO;
               *locLij = *locUjk = 0;
            }                   /* end if */
         }                      /* end if lj > ui */
      }                         /* end while search == YES */
   }                            /* end if (*locLij <= endLix) && (*locUjk <=
                                 * endUxk) */
   else
   {
      *locLij = *locUjk = 0;
   }                            /* end else */

}                               /* end locLijUjk() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateLxk(k)
int             k;

{
   int             i;
   int             locLik, locLij, locUjk;
   int             endLix, endUxk;
   complex         c1;

   locLik = diag[k].NextLxk;
   while (locLik != 0)
   {
      i = blockrow[locLik].i - blocks[0].DiagPtr;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\tblockrow[%d].i=%d - blockrow[%d].j=%d - i=%d \n",
        locLik,blockrow[locLik].i,locLik,blockrow[locLik].j,i);
   fflush(stdout); }
*/
/* initialize the pointers into the data structures */
      locLij = (diag[(i - 1)].PtrLix + 1);
      locUjk = (diag[(k - 1)].PtrUxj + 1);
      endLix = diag[i].PtrLix;
      endUxk = diag[k].PtrUxj;
/*
printf("\t\t\t1 - locLij=%d - endLix=%d - locUjk=%d - endUxk=%d - k=%d \n",
            locLij,endLix,locUjk,endUxk,k);
fflush(stdout);
*/
/* find the first pair of nonzero Lik and Ukj */
      locLijUjk(k, &locLij, &locUjk, endLix, endUxk);
/*
printf("\t\t\t2 - locLij=%d - endLix=%d - locUjk=%d - endUxk=%d - k=%d \n",
            locLij,endLix,locUjk,endUxk,k);
fflush(stdout);
*/
      while ((locLij != 0) && (locUjk != 0))
      {

         /*
          * blockrow[locLik].Value -= blockrow[locLij].Value *
          * blockcol[locUjk].Value;
          */
         multiply(blockrow[locLij].Value, blockcol[locUjk].Value, c1);
         subtract(blockrow[locLik].Value, c1, blockrow[locLik].Value);

         locLij++;
         locUjk++;
         nflopsl++;
/* find the next pair of nonzero Lij and Ujk */
         locLijUjk(k, &locLij, &locUjk, endLix, endUxk);
/*
printf("\t\t\t3 - locLij=%d - endLix=%d - locUjk=%d - endUxk=%d - k=%d \n",
            locLij,endLix,locUjk,endUxk,k);
fflush(stdout);
*/
      }                         /* end while (locLij and locUjk != 0) */
      locLik = blockrow[locLik].NextLxk;
   }                            /* end while (locLik !=0) */
}                               /* end UpdateLxk() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateLBxk(k, l)
int             k;

int             l;

{
   int             i;
   int             locLBik, locLBij, locUjk;
   int             endLBix, endUxk;
   int             LBPtr, EndLBPtr, LBix;
   complex         c1;

/* initialize the pointers into the data structures */
   LBPtr = (blocks[l - 1].LBPtr + 1);
   EndLBPtr = blocks[l].LBPtr;
   LBix = lborder[LBPtr].i;

   locLBik = diag[k].NextLBxk;
   while (locLBik != 0)
   {
      i = borderrow[locLBik].i;
/* find the location of the ith row in the lower border */
      while ((LBix < i) && (LBPtr < EndLBPtr))
         LBix = lborder[++LBPtr].i;
      if (LBix != i)
      {
         printf("LBix != i in UpdateLBxk(%d,%d) LBix=%d i=%d\n", k, l, LBix, i);
         exit(1);
      }                         /* end if LBix != i */

      locLBij = (lborder[(LBPtr - 1)].PtrLBix + 1);
      locUjk = (diag[(k - 1)].PtrUxj + 1);
      endLBix = lborder[LBPtr].PtrLBix;
      endUxk = diag[k].PtrUxj;
/* find the first pair of nonzero LBij and Ujk */
      locLBijUjk(k, &locLBij, &locUjk, endLBix, endUxk);
      while ((locLBij != 0) && (locUjk != 0))
      {

         /*
          * borderrow[locLBik].Value -= borderrow[locLBij].Value *
          * blockcol[locUjk].Value;
          */
         multiply(borderrow[locLBij].Value, blockcol[locUjk].Value, c1);
         subtract(borderrow[locLBik].Value, c1, borderrow[locLBik].Value);

         locLBij++;
         locUjk++;
         nflopsl++;
/* find the next pair of nonzero LBij and Ujk */
         locLBijUjk(k, &locLBij, &locUjk, endLBix, endUxk);
      }                         /* end while (locLBij and locUjk != 0) */
      locLBik = borderrow[locLBik].NextLBxk;
   }                            /* end while (locLBik !=0) */

}                               /* end UpdateLBik() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            locLBijUjk(k, locLBij, locUjk, endLBix, endUxk)
int             k, *locLBij, *locUjk, endLBix, endUxk;

{
   int             Search;
   int             lj, ui;

   if ((*locLBij <= endLBix) && (*locUjk <= endUxk))
   {
      lj = borderrow[*locLBij].j - blocks[0].DiagPtr;
      ui = blockcol[*locUjk].i - blocks[0].DiagPtr;
      Search = YES;
      while (Search == YES)
      {
         if (lj == ui)
         {
            Search = NO;
         }                      /* end if lj == ui */
         else if (lj < ui)
         {
            while ((lj < ui) && (lj < k) && (*locLBij <= endLBix))
               lj = borderrow[++(*locLBij)].j - blocks[0].DiagPtr;
            if ((lj >= k) || (*locLBij > endLBix))
            {
               Search = NO;
               *locLBij = *locUjk = 0;
            }                   /* end if */
         }                      /* end if lj < ui */
         else if (lj > ui)
         {
            while ((lj > ui) && (ui < k) && (*locUjk <= endUxk))
               ui = blockcol[++(*locUjk)].i - blocks[0].DiagPtr;
            if ((ui >= k) || (*locUjk > endUxk))
            {
               Search = NO;
               *locLBij = *locUjk = 0;
            }                   /* end if */
         }                      /* end if lj > ui */
      }                         /* end while search == YES */
   }                            /* end if (*locLBij <= endLBix) && (*locUjk
                                 * <= endUxk) */
   else
   {
      *locLBij = *locUjk = 0;
   }                            /* end else */

}                               /* end locLBijUjk() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateUkx(k)
int             k;

{
   int             j;
   int             locUkj, locLki, locUij;
   int             endLkx, endUxj;
   complex         c1;

   locUkj = diag[k].NextUkx;
   while (locUkj != 0)
   {
      j = blockcol[locUkj].j - blocks[0].DiagPtr;
/* initialize the pointers into the data structures */
      locLki = (diag[(k - 1)].PtrLix + 1);
      locUij = (diag[(j - 1)].PtrUxj + 1);
      endLkx = diag[k].PtrLix;
      endUxj = diag[j].PtrUxj;
/* find the first pair of nonzero Lki and Uij */
      locLijUjk(k, &locLki, &locUij, endLkx, endUxj);
      while ((locLki != 0) && (locUij != 0))
      {

         /*
          * blockcol[locUkj].Value -= blockrow[locLki].Value *
          * blockcol[locUij].Value;
          */
         multiply(blockrow[locLki].Value, blockcol[locUij].Value, c1);
         subtract(blockcol[locUkj].Value, c1, blockcol[locUkj].Value);

         locLki++;
         locUij++;
         nflopsu++;
/* find the next pair of nonzero Lki and Uij */
         locLijUjk(k, &locLki, &locUij, endLkx, endUxj);
      }                         /* end while (locLki and locUij != 0) */
/* divide the value of Ukj by Lkk */
      /* blockcol[locUkj].Value /= diag[k].Value; */
      division(blockcol[locUkj].Value, diag[k].Value, blockcol[locUkj].Value);

      locUkj = blockcol[locUkj].NextUkx;
   }                            /* end while (locUkj !=0) */

}                               /* end UpdateUkx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateUBkx(k, l)
int             k;

int             l;

{
   int             j;
   int             locUBkj, locLki, locUBij;
   int             endLkx, endUBxj;
   int             UBPtr, EndUBPtr, UBxj;
   complex         c1;

/* initialize the pointers into the data structures */
   UBPtr = (blocks[l - 1].UBPtr + 1);
   EndUBPtr = blocks[l].UBPtr;
   UBxj = uborder[UBPtr].j;


   locUBkj = diag[k].NextUBkx;
   while (locUBkj != 0)
   {
      j = bordercol[locUBkj].j;
/* find the location of the jth column in the upper border */
      while ((UBxj < j) && (UBPtr < EndUBPtr))
         UBxj = uborder[++UBPtr].j;
      if (UBxj != j)
      {
         printf("UBxj != j in UpdateUBxj(%d,%d) UBxj=%d j=%d\n", k, l, UBxj, j);
         exit(1);
      }                         /* end if UBxj != j */

      locLki = (diag[(k - 1)].PtrLix + 1);
      locUBij = (uborder[(UBPtr - 1)].PtrUBxj + 1);
      endLkx = diag[k].PtrLix;
      endUBxj = uborder[UBPtr].PtrUBxj;
/* find the first pair of nonzero Lkx and UBxj */
      locLkiUBij(k, &locLki, &locUBij, endLkx, endUBxj);
      while ((locLki != 0) && (locUBij != 0))
      {

         /*
          * bordercol[locUBkj].Value -= blockrow[locLki].Value *
          * bordercol[locUBij].Value;
          */
         multiply(blockrow[locLki].Value, bordercol[locUBij].Value, c1);
         subtract(bordercol[locUBkj].Value, c1, bordercol[locUBkj].Value);

         locLki++;
         locUBij++;
         nflopsu++;
/* find the next pair of nonzero Lix and UBxj */
         locLkiUBij(k, &locLki, &locUBij, endLkx, endUBxj);
      }                         /* end while (locLki and locUBij != 0) */
/* divide the value of UBkj by Lkk */
      /* bordercol[locUBkj].Value /= diag[k].Value; */
      division(bordercol[locUBkj].Value, diag[k].Value, bordercol[locUBkj].Value);

      locUBkj = bordercol[locUBkj].NextUBkx;
   }                            /* end while (locUBkj !=0) */

}                               /* end UpdateUBkx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            locLkiUBij(k, locLki, locUBij, endLix, endUBxj)
int             k, *locLki, *locUBij, endLix, endUBxj;

{
   int             Search;
   int             lj, ui;

   if ((*locLki <= endLix) && (*locUBij <= endUBxj))
   {
      lj = blockrow[*locLki].j - blocks[0].DiagPtr;
      ui = bordercol[*locUBij].i - blocks[0].DiagPtr;
      Search = YES;
      while (Search == YES)
      {
         if (lj == ui)
         {
            Search = NO;
         }                      /* end if lj == ui */
         else if (lj < ui)
         {
            while ((lj < ui) && (lj < k) && (*locLki <= endLix))
               lj = blockrow[++(*locLki)].j - blocks[0].DiagPtr;
            if ((lj >= k) || (*locLki > endLix))
            {
               Search = NO;
               *locLki = *locUBij = 0;
            }                   /* end if */
         }                      /* end if lj < ui */
         else if (lj > ui)
         {
            while ((lj > ui) && (ui < k) && (*locUBij <= endUBxj))
               ui = bordercol[++(*locUBij)].i - blocks[0].DiagPtr;
            if ((ui >= k) || (*locUBij > endUBxj))
            {
               Search = NO;
               *locLki = *locUBij = 0;
            }                   /* end if */
         }                      /* end if lj > ui */
      }                         /* end while search == YES */
   }                            /* end if (*locLki <= endLix) && (*locUBij <=
                                 * endUBxj) */
   else
   {
      *locLki = *locUBij = 0;
   }                            /* end else */

}                               /* end locLkiUBij() */

#if CMAML_RPC
/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateLastBlockCMAML()
{
   int             i, j, k, l, p, i0, i1, i2, i3;
   int             iproc;
   int             LBPtr, EndLBPtr;
   int             UBPtr, EndUBPtr;
   int             locLBij, locUBjk;
   int             endLBix, endUBxk;
   int             locLij, endLij, locUij, endUij;
   unsigned int    iupd1, iupd2;
   complex         update, iupd, c1;

/*
printf("\tCalculate and sum the updates to the last block \n");
fflush(stdout);
*/
/* calculate and sum the updates to the last block
   this subroutine calculates the product of two
   sparse matrices  */
/*
printf("In UpdateLastBlockCMAML() - nprocs=%d - nupdatesULB[%d]=%d \n",nprocs,(procnum+1),nupdatesULB[procnum+1]);
fflush(stdout);
*/
   ipollULBCMAML = 0;
   for (l = 1; l <= nblocks; l++)
   {
/*
printf("\t\tl=%d - nblocks=%d \n",l,nblocks);
fflush(stdout);
*/
/* loop over all non-zero lower border rows in this block ---
   also visit all non-zero upper border columns in the corresponding
   block location --- visit them in a pair of nested while loops */
/* outer loop */
      for (LBPtr = (blocks[l - 1].LBPtr + 1); LBPtr <= blocks[l].LBPtr; LBPtr++)
      {
         i = lborder[LBPtr].i - lastblockstart + 1;
/* inner loop */
         for (UBPtr = (blocks[l - 1].UBPtr + 1); UBPtr <= blocks[l].UBPtr; UBPtr++)
         {
            update.real = 0.0;
            update.imag = 0.0;
            locLBij = (lborder[LBPtr - 1].PtrLBix + 1);
            endLBix = lborder[LBPtr].PtrLBix;
            j = uborder[UBPtr].j - lastblockstart + 1;
            locUBjk = (uborder[UBPtr - 1].PtrUBxj + 1);
            endUBxk = uborder[UBPtr].PtrUBxj;
/* data must exist in this partial row and partial column */
/* find the first pair of nonzero LBij and UBjk */
            locLBijUBjk(&locLBij, &locUBjk, endLBix, endUBxk);
            while ((locLBij != 0) && (locUBjk != 0))
            {

               /*
                * update += borderrow[locLBij].Value *
                * bordercol[locUBjk].Value;
                */
               multiply(borderrow[locLBij].Value, bordercol[locUBjk].Value, c1);
               add(update, c1, update);

               if (borderrow[locLBij].i == bordercol[locUBjk].j)
                  nflopsd++;
               else if (borderrow[locLBij].i > bordercol[locUBjk].j)
                  nflopsl++;
               else
                  nflopsu++;

               locLBij++;
               locUBjk++;
/*
printf("locLBij=%d - locUBj=%d - nflops=%d -  \n",locLBij,locUBj,nflops);
fflush(stdout);
*/
/* find the next pair of nonzero LBij and UBjk */
               locLBijUBjk(&locLBij, &locUBjk, endLBix, endUBxk);
            }                   /* end while (locLBij and locUBjk != 0) */
            i0 = i - 1;
            iproc = ((i0 % (nprocs * nb)) / nb);
/*
printf("\tiproc=%d - nb=%d - i0=%d - i=%d - j=%d \n",iproc,nb,i0,i,j);
fflush(stdout);
*/
            if (iproc != procnum)
            {
               iupd1 = (int) update.real;
               iupd2 = (int) update.imag;
               CMAML_rpc(iproc, StoreHandlerULBCMAML, i, j, iupd1, iupd2);
            }                   /* end if */
            else
            {
               i1 = (i0 / (nb * nprocs));
               i2 = (i0 % nb);
               i3 = (i1 * nb) + i2 + 1;
/*
printf("\t\ti=%d - i0=%d - i1=%d - i2=%d - i3=%d \n",i,i0,i1,i2,i3);
fflush(stdout);
*/
               if (i > j)
               {
                  locLij = (lbdiag[i3 - 1].PtrLix + 1);
                  endLij = lbdiag[i3].PtrLix;

                  /* the value Lij must exist so search for it */
                  while ((lbrowL[locLij].j < j) && (locLij < endLij))
                     locLij++;
                  if ((locLij > endLij) || (lbrowL[locLij].j != j))
                  {
                     CMMD_error("error in UpdateLastBlockCMAML() --- (locLij > endLij) - locLij=%d -  endLij=%d - i=%d - j=%d \n",
                        locLij, endLij, i, j);
                  }
/*
printf("\t\t\t(i > j) - i3=%d - locLij=%d - endLij=%d \n",i3,locLij,endLij);
fflush(stdout);
*/
                  /* lbrowL[locLij].Value -= update; */
                  subtract(lbrowL[locLij].Value, update, lbrowL[locLij].Value);
               }                /* end else if i > j */

               else if (i == j)
               {
/*
printf("\t\t\t(i = j) - i3=%d \n",i3);
fflush(stdout);
*/
                  /* lbdiag[i3].Value -= update; */
                  subtract(lbdiag[i3].Value, update, lbdiag[i3].Value);
               }                /* end if i == j */
               else
               {
                  /* if (i < j) */
                  locUij = (lbdiag[i3 - 1].PtrUix + 1);
                  endUij = lbdiag[i3].PtrUix;

                  /* the value Uij must exist so search for it */
                  while ((lbrowU[locUij].j < j) && (locUij < endUij))
                     locUij++;
                  if ((locUij > endUij) || (lbrowU[locUij].j != j))
                  {
                     CMMD_error("error in UpdateLastBlockCMAML() --- (locUij > endUij) - locUij=%d -  endUij=%d - i=%d - j=%d \n",
                        locUij, endUij, i, j);
                  }
/*
printf("\t\t\t(i < j) - i3=%d - locUij=%d - endUij=%d \n",i3,locUij,endUij);
fflush(stdout);
*/
                  /* lbrowU[locUij].Value -= update; */
                  subtract(lbrowU[locUij].Value, update, lbrowU[locUij].Value);
               }                /* end else if i < j */

               ipollULBCMAML++;
            }                   /* end else */
            CMAML_poll();
/*
printf("======>>>> ipollULBCMAML=%d - nupdatesULB[%d]=%d \n",
                   ipollULBCMAML,(procnum+1),nupdatesULB[procnum+1]);
fflush(stdout);
*/
         }                      /* end inner loop */
      }                         /* end outer loop */
   }                            /* end for l */
/* check to see if data has been received */
/*
printf("======>>>> before CMAML_poll_while_lt - ipollULBCMAML=%d - nupdatesULB[%d]=%d \n",
                   ipollULBCMAML,(procnum+1),nupdatesULB[procnum+1]);
fflush(stdout);
*/
   CMAML_poll_while_lt(&ipollULBCMAML, nupdatesULB[procnum + 1]);
/*
printf("Exit GSUpdateLastBlockCMAML() - ipollULBCMAML=%d - nupdatesULB[%d]=%d \n",
           ipollULBCMAML,(procnum+1),nupdatesULB[procnum+1]);
fflush(stdout);
*/
}                               /* end UpdateLastBlockCMAML() */

#endif

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            StoreHandlerULBCMAML(i, j, iupd1, iupd2)
int             i, j;

unsigned int    iupd1, iupd2;

{
   int             i0, i1, i2, i3, locLij, endLij, locUij, endUij;
   complex         iupd;

   iupd.real = (float) iupd1;
   iupd.imag = (float) iupd2;

   i0 = i - 1;
   i1 = (i0 / (nb * nprocs));
   i2 = (i0 % nb);
   i3 = (i1 * nb) + i2 + 1;

   if (i > j)
   {
      locLij = (lbdiag[i3 - 1].PtrLix + 1);
      endLij = lbdiag[i3].PtrLix;

      /* the value Lij must exist so search for it */
      while ((lbrowL[locLij].j < j) && (locLij < endLij))
         locLij++;
      if ((locLij > endLij) || (lbrowL[locLij].j != j))
      {
         CMMD_error("%d - error in UpdateLastBlockCMAML() --- (locLij > endLij) - locLij=%d -  endLij=%d - i=%d - j=%d - i3=%d - lbrowL[%d].j=%d \n",
            procnum, locLij, endLij, i, j, i3, locLij, lbrowL[locLij].j);
      }
      /* lbrowL[locLij].Value -= * ((double *) (&iupd)); */
      subtract(lbrowL[locLij].Value, iupd, lbrowL[locLij].Value);
   }                            /* end else if i > j */

   else if (i == j)
   {
      /* lbdiag[i3].Value -= * ((double *) (&iupd)); */
      subtract(lbdiag[i3].Value, iupd, lbdiag[i3].Value);
   }                            /* end if i == j */
   else
   {
      /* if (i < j) */
      locUij = (lbdiag[i3 - 1].PtrUix + 1);
      endUij = lbdiag[i3].PtrUix;

      /* the value Uij must exist so search for it */
      while ((lbrowU[locUij].j < j) && (locUij < endUij))
         locUij++;
      if ((locUij > endUij) || (lbrowU[locUij].j != j))
      {
         CMMD_error("%d - error in UpdateLastBlockCMAML() --- (locUij > endUij) - locUij=%d -  endUij=%d - i=%d - j=%d - i3=%d \n",
            procnum, locUij, endUij, i, j, i3);
      }
      /* lbrowU[locUij].Value -= * ((double *) (&iupd)); */
      subtract(lbrowU[locUij].Value, iupd, lbrowU[locUij].Value);
   }                            /* end else if i < j */

   ipollULBCMAML++;
}                               /* end StoreHandlerULBCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            locLBijUBjk(locLBij, locUBjk, endLBix, endUBxk)
int            *locLBij, *locUBjk, endLBix, endUBxk;

{
   int             Search;
   int             lj, ui;

   if ((*locLBij <= endLBix) && (*locUBjk <= endUBxk))
   {
      lj = borderrow[*locLBij].j;
      ui = bordercol[*locUBjk].i;
      Search = YES;
      while (Search == YES)
      {
         if (lj == ui)
         {
            Search = NO;
         }                      /* end if lj == ui */
         else if (lj < ui)
         {
            while ((lj < ui) && (*locLBij <= endLBix))
               lj = borderrow[++(*locLBij)].j;
            if (*locLBij > endLBix)
            {
               Search = NO;
               *locLBij = *locUBjk = 0;
            }                   /* end if */
         }                      /* end if lj < ui */
         else if (lj > ui)
         {
            while ((lj > ui) && (*locUBjk <= endUBxk))
               ui = bordercol[++(*locUBjk)].i;
            if (*locUBjk > endUBxk)
            {
               Search = NO;
               *locLBij = *locUBjk = 0;
            }                   /* end if */
         }                      /* end if lj > ui */
      }                         /* end while search == YES */
   }                            /* end if (*locLBij <= endLBix) && (*locUBjk
                                 * <= endUBxk) */
   else
   {
      *locLBij = *locUBjk = 0;
   }                            /* end else */

}                               /* end locLBijUBjk() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            UpdateLastBlock()
{
   int             i, j, k, l, p, i0, i1, i2, i3, loc, bufsize;
   int             iproc, nextproc, prevproc, UplbStatus;
   int             LBPtr, EndLBPtr;
   int             UBPtr, EndUBPtr;
   int             locLBij, locUBjk;
   int             endLBix, endUBxk;
   int             locLij, endLij, locUij, endUij;
   complex         update, c1;
   CMMD_mcb        UplbMCB;

#ifdef DEBUG_UpdateLastBlock
   printf("\tCalculate and sum the updates to the last block \n");
   fflush(stdout);
#endif

/* calculate and sum the updates to the last block */
/* calculate and sum the updates to the last block
   this subroutine calculates the product of two
   sparse matrices  */

#ifdef DEBUG_UpdateLastBlock
   printf("In UpdateLastBlock() - nprocs=%d - nupdatesULB[%d]=%d \n",
      nprocs, (procnum + 1), nupdatesULB[procnum + 1]);
   fflush(stdout);
#endif

   for (l = 1; l <= nblocks; l++)
   {

#ifdef DEBUG_UpdateLastBlock
      printf("\t\tl=%d - nblocks=%d \n", l, nblocks);
      fflush(stdout);
#endif

/* loop over all non-zero lower border rows in this block ---
   also visit all non-zero upper border columns in the corresponding
   block location --- visit them in a pair of nested while loops */
/* outer loop */
      for (LBPtr = (blocks[l - 1].LBPtr + 1); LBPtr <= blocks[l].LBPtr; LBPtr++)
      {
         i = lborder[LBPtr].i - lastblockstart + 1;
/* inner loop */
         for (UBPtr = (blocks[l - 1].UBPtr + 1); UBPtr <= blocks[l].UBPtr;
            UBPtr++)
         {
            update.real = 0.0;
            update.imag = 0.0;
            locLBij = (lborder[LBPtr - 1].PtrLBix + 1);
            endLBix = lborder[LBPtr].PtrLBix;
            j = uborder[UBPtr].j - lastblockstart + 1;
            locUBjk = (uborder[UBPtr - 1].PtrUBxj + 1);
            endUBxk = uborder[UBPtr].PtrUBxj;
/* data must exist in this partial row and partial column */
/* find the first pair of nonzero LBij and UBjk */
            locLBijUBjk(&locLBij, &locUBjk, endLBix, endUBxk);
            while ((locLBij != 0) && (locUBjk != 0))
            {

               /*
                * update += borderrow[locLBij].Value *
                * bordercol[locUBjk].Value;
                */
               multiply(borderrow[locLBij].Value, bordercol[locUBjk].Value, c1);
               add(update, c1, update);

               locLBij++;
               locUBjk++;
               nflops++;
/* find the next pair of nonzero LBij and UBjk */
               locLBijUBjk(&locLBij, &locUBjk, endLBix, endUBxk);
            }                   /* end while (locLBij and locUBjk != 0) */
            i0 = i - 1;
            iproc = ((i0 % (nprocs * nb)) / nb);
            if (iproc != procnum)
            {
               lbupdate[iproc][nlbupdate[iproc]].Value.real = update.real;
               lbupdate[iproc][nlbupdate[iproc]].Value.imag = update.imag;
               lbupdate[iproc][nlbupdate[iproc]].i = i;
               lbupdate[iproc][nlbupdate[iproc]].j = j;
               (nlbupdate[iproc])++;
            }                   /* end if */
            else
            {
               i1 = (i0 / (nb * nprocs));
               i2 = (i0 % nb);
               i3 = (i1 * nb) + i2 + 1;

               if (i > j)
               {
                  locLij = (lbdiag[i3 - 1].PtrLix + 1);
                  endLij = lbdiag[i3].PtrLix;

                  /* the value Lij must exist so search for it */
                  while ((lbrowL[locLij].j < j) && (locLij < endLij))
                     locLij++;
                  if ((locLij > endLij) || (lbrowL[locLij].j != j))
                  {
                     CMMD_error("error in UpdateLastBlockCMAML() --- (locLij > endLij) - locLij=%d -  endLij=%d - i=%d - j=%d \n",
                        locLij, endLij, i, j);
                  }
                  /* lbrowL[locLij].Value -= update; */
                  subtract(lbrowL[locLij].Value, update, lbrowL[locLij].Value);
               }                /* end else if i > j */

               else if (i == j)
               {
                  /* lbdiag[i3].Value -= update; */
                  subtract(lbdiag[i3].Value, update, lbdiag[i3].Value);
               }                /* end if i == j */

               else
               {
                  /* if (i < j) */
                  locUij = (lbdiag[i3 - 1].PtrUix + 1);
                  endUij = lbdiag[i3].PtrUix;

                  /* the value Uij must exist so search for it */
                  while ((lbrowU[locUij].j < j) && (locUij < endUij))
                     locUij++;
                  if ((locUij > endUij) || (lbrowU[locUij].j != j))
                  {
                     CMMD_error("error in UpdateLastBlockCMMD() --- (locUij > endUij) - locUij=%d -  endUij=%d - i=%d - j=%d \n",
                        locUij, endUij, i, j);
                  }
                  /* lbrowU[locUij].Value -= update; */
                  subtract(lbrowU[locUij].Value, update, lbrowU[locUij].Value);
               }                /* end else if i < j */

            }                   /* end else */
         }                      /* end inner loop */
      }                         /* end outer loop */
   }                            /* end for l */

/* the update buffer only has nonzero values - so mark the
   last value with zeros in the row and column markers */
   for (iproc = 0; iproc < nprocs; iproc++)
   {
      lbupdate[iproc][nlbupdate[iproc]].Value.real = 0.0;
      lbupdate[iproc][nlbupdate[iproc]].Value.imag = 0.0;
      lbupdate[iproc][nlbupdate[iproc]].i = 0;
      lbupdate[iproc][nlbupdate[iproc]].j = 0;

#ifdef DEBUG_UpdateLastBlock
      printf("\t\tnlbupdate[%d]=%d \n", iproc, nlbupdate[iproc]);
      fflush(stdout);
#endif
   }                            /* end for iproc */

/* update the values in the last block  - this is the portion
   of the Shur complement that updates the last block using
   the values in the borders */

/* send the appropriate data to the other processors - each processor
   will send data asychronously to another processor equidistant away -
   until all processors have talked to all other processsors */

   nextproc = procnum;
   prevproc = procnum;
   for (iproc = 0; iproc < (nprocs - 1); iproc++)
   {
      nextproc = ((nextproc + 1) % nprocs);
      prevproc = ((prevproc + nprocs - 1) % nprocs);
      bufsize = ((nlbupdate[nextproc] + 1) * sizeof(LBRow));

#ifdef DEBUG_UpdateLastBlock
      printf("\t\t\tiproc=%d - nextproc=%d - prevproc=%d - nlbupdate[%d]=%d - (bytes)=%d \n",
         iproc, nextproc, prevproc, nextproc, nlbupdate[nextproc], bufsize);
      fflush(stdout);
#endif

#ifdef CMMD
#if 0
#define DEBUG_UpdateLastBlock
#endif
#ifdef DEBUG_UpdateLastBlock
      printf("%d UpdateLastBlock send dest %d bufsize %d\n",
         CMMD_self_address(), nextproc, bufsize);
#endif
      CMMD_send_async(nextproc, (SEND_UPDATELBFAC + procnum),
         lbupdate[nextproc], bufsize, SendHandlerCMMD, NULL);

      bufsize = lbupdatercv_size;
#if 0
#define DEBUG_UpdateLastBlock
#endif
#ifdef DEBUG_UpdateLastBlock
      printf("%d UpdateLastBlock recv src %d bufsize %d\n",
         CMMD_self_address(), prevproc, bufsize);
#endif
      UplbMCB = CMMD_receive_async(prevproc, (RECEIVE_UPDATELBFAC + prevproc),
         lbupdatercv, bufsize, ReceiveHandlerCMMD,
         &UplbStatus);
/* check to see if data has been received */
      if (UplbStatus != (RECEIVE_UPDATELBFAC + prevproc))
      {

#ifdef DEBUG_UpdateLastBlock
         printf("Wait to receive x[i] from processor - %d ---  \n", prevproc);
         fflush(stdout);
#endif

         CMMD_msg_wait(UplbMCB);
      }                         /* end if */
#endif

      k = 0;
      while ((lbupdatercv[k].i != 0) && (lbupdatercv[k].j != 0))
      {
         i = lbupdatercv[k].i;
         j = lbupdatercv[k].j;
         i0 = i - 1;
         i1 = (i0 / (nb * nprocs));
         i2 = (i0 % nb);
         i3 = (i1 * nb) + i2 + 1;

         if (i > j)
         {
            locLij = (lbdiag[i3 - 1].PtrLix + 1);
            endLij = lbdiag[i3].PtrLix;

            /* the value Lij must exist so search for it */
            while ((lbrowL[locLij].j < j) && (locLij <= endLij))
               locLij++;
            if (locLij > endLij)
            {
               CMMD_error("error in UpdateLastBlockCMAML() --- (locLij > endLij) - locLij=%d -  endLij=%d - i=%d - j=%d \n",
                  locLij, endLij, i, j);
            }
            /* lbrowL[locLij].Value -= lbupdatercv[k].Value; */
            subtract(lbrowL[locLij].Value, lbupdatercv[k].Value, lbrowL[locLij].Value);
         }                      /* end else if i > j */

         else if (i == j)
         {
            /* lbdiag[i3].Value -= lbupdatercv[k].Value; */
            subtract(lbdiag[i3].Value, lbupdatercv[k].Value, lbdiag[i3].Value);
         }                      /* end if i == j */

         else
         {
            /* if (i < j) */
            locUij = (lbdiag[i3 - 1].PtrUix + 1);
            endUij = lbdiag[i3].PtrUix;

            /* the value Uij must exist so search for it */
            while ((lbrowU[locUij].j < j) && (locUij <= endUij))
               locUij++;
            if (locUij > endUij)
            {
               CMMD_error("error in UpdateLastBlockCMAML() --- (locUij > endUij) - locUij=%d -  endUij=%d - i=%d - j=%d \n",
                  locUij, endUij, i, j);
            }
            /* lbrowU[locUij].Value -= lbupdatercv[k].Value; */
            subtract(lbrowU[locUij].Value, lbupdatercv[k].Value, lbrowU[locUij].Value);
         }                      /* end else if i < j */

         k++;
      }                         /* end while */
   }                            /* end for iproc */
#if 0
#define DEBUG_UpdateLastBlock
#endif
#ifdef DEBUG_UpdateLastBlock
   printf("end UpdateLastBlock()\n");
#undef DEBUG_UpdateLastBlock
#endif
}                               /* end UpdateLastBlock() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SendHandlerCMMD(mcb, arg)
CMMD_mcb       *mcb;

void           *arg;

{
   CMMD_free_mcb(*mcb);
}                               /* end SendHandlerCMMD() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ReceiveHandlerCMMD(mcb, Status)
CMMD_mcb       *mcb;

int            *Status;

{
   *Status = CMMD_mcb_tag(*mcb);
   CMMD_free_mcb(*mcb);
}                               /* end ReceiveHandlerCMMD() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            FactorLastBlock()
{

   BlockSparseFac();

}                               /* end FactorLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
/* block LU factorization for sparse matrices with
   no inherient parallelism - no pivoting will be performed
   and all fillin locations are known a priori

  This code was inspired by hlltbk.f authored by Manish Parashar of NPAC */

void            BlockSparseFac()
{
/*
   nlb          is the number of rows in the matrix  (global)
   nb           is the number of rows in a block     (global)
   nprocs       is the number of processors          (global)
   rowsthisproc is the number of rows this processor (global)
*/
   int             i, j, k, jb, jb1, origproc, icomm;
   int             rowsfin, npanels, nlbrowUsend, nlbrowUrcv;
   int             lbrowUloc, bufsize;

   npanels = rowsfin = origproc = icomm = 0;

/* for one processor --- nprocs == 1 */
   if (nprocs == 1)             /* use unblocked, sequential code on
                                 * processor 0 */
   {
/*
printf("for one processor --- nb=%d - nlb=%d - procnum=%d \n",nb,nlb,procnum);
fflush(stdout);
*/
      SparsekijLUSeq();
   }                            /* end if one processor */
   else
   {
      /* use blocked, parallel code on processors less than nprocs */

#if 0
#define DEBUG_BlockSparseFac
#endif

#ifdef DEBUG_BlockSparseFac
      printf("for multiple  processors --- nb=%d - nlb=%d - procnum=%d \n",
         nb, nlb, procnum);
      fflush(stdout);
#undef DEBUG_BlockSparseFac
#endif

      if (procnum == 0)         /* factor the first block and send the values */
      {
         SparsekijLUPar(1, nb);
         nlbrowUsend = lbdiag[nb].PtrUix;
         lbrowUloc = 1;
         bufsize = (nlbrowUsend * sizeof(LBRow));
         activesendSLB(&(lbrowU[lbrowUloc].Value), 'N', bufsize);
/*
printf("\tfactor the first block and send the values --- nlbrowUsend=%d - procnum=%d \n",nlbrowUsend,procnum);
fflush(stdout);
*/
      }                         /* end if procnum == 0 */

      for (j = 1; j <= (nlb - nb); (j = j + nb))
      {
         jb = ((nlb - j + 1) < nb) ? (nlb - j + 1) : nb;        /* jb=min( (nlb-j+1) ,
                                                                 * nb ) */
/*
printf("----->>>>> j=%d - nlb=%d - nb=%d -- jb=%d <<<<<----- \n",j,nlb,nb,jb);
fflush(stdout);
*/
         if (origproc == procnum)       /* note the rows updated in the
                                         * previous step */
         {
            npanels++;
            rowsfin += jb;
            icomm++;
/*
printf("\t\torigproc=%d - procnum=%d - npanels=%d - rowsfin=%d - jb=%d \n",
        origproc,procnum,npanels,rowsfin,jb);
fflush(stdout);
*/
         }                      /* end if origproc == procnum */
         else
         {
            /* receive a panel */
            nlbrowUrcv = lbbufsizeFac[++icomm];
            bufsize = (nlbrowUrcv * sizeof(LBRow));
/*
if (procnum == 0) {
  printf("\treceive a panel --- nlbrowUrcv=%d - origproc=%d - procnum=%d - npanels=%d - rowsfin=%d - bufsize=%d - icomm=%d \n",
                    nlbrowUrcv,origproc,procnum,npanels,rowsfin,bufsize,icomm);
  fflush(stdout); }
*/
            activerecvSLB(lbrowUrcv, 'N', bufsize, origproc);
         }                      /* end else */
/*
printf("\t\t  ==>>   origproc=%d - procnum=%d - npanels=%d - rowsfin=%d - nprocs=%d \n",
        origproc,procnum,npanels,rowsfin,nprocs);
fflush(stdout);
*/
         origproc = ((origproc + 1) % nprocs);
/*
if (procnum == 7) {
  printf("\t\t====>>>> origproc=%d - procnum=%d - npanels=%d - rowsfin=%d - nprocs=%d - rowsthisproc=%d - (rowsthisproc-rowsfin)=%d \n",
        origproc,procnum,npanels,rowsfin,nprocs,rowsthisproc,(rowsthisproc-rowsfin));
  fflush(stdout); }
*/
         if ((rowsthisproc - rowsfin) > 0)
         {
            if (origproc == procnum)
            {
               jb1 = ((nlb - (j + nb) + 1) < nb) ? (nlb - (j + nb) + 1) : nb;   /* jb1 =
                                                                                 * min(n-(j+nb)+1,nb) */
/*
printf("----->>>>> j=%d - nlb=%d - nb=%d - jb1=%d <<<<<----- \n",j,nlb,nb,jb1);
fflush(stdout);
*/
               /* this processor is the next to factor a panel */
               /* update only the first panel */
/*
if (procnum == 7) {
  printf("\t\t update only the first panel --- rowsfin=%d - jb1=%d - nlbrowUrcv=%d - npanels=%d - origproc=%d \n",
                    rowsfin,jb1,nlbrowUrcv,npanels,origproc);
  fflush(stdout); }
*/
               SparsekijLUUpdate((rowsfin + 1), (rowsfin + jb1), lbrowUrcv,
                                 nlbrowUrcv);
               /* factor the first panel and send it */
/*
printf("\t\tfactor a panel --- rowsfin=%d - jb=%d \n",rowsfin,jb);
fflush(stdout);
*/
               SparsekijLUPar((rowsfin + 1), jb1);
               if (j <= (nlb - (2 * nb)))       /* send all but last block -
                                                 * avoid a null send */
               {
                  nlbrowUsend = lbdiag[rowsfin + jb1].PtrUix - lbdiag[rowsfin].PtrUix;
                  lbrowUloc = lbdiag[rowsfin].PtrUix + 1;
                  bufsize = (nlbrowUsend * sizeof(LBRow));
/*
printf("\t\tsend a panel --- nlbrowUsend=%d - rowsfin=%d - j=%d - lbrowUloc=%d - bufsize=%d - icomm=%d - jb1=%d \n",
        nlbrowUsend,rowsfin,j,lbrowUloc,bufsize,icomm,jb1);
fflush(stdout);
*/
                  activesendSLB(&(lbrowU[lbrowUloc]), 'N', bufsize);
               }                /* end if */
            }                   /* end if origproc == procnum */

/* update the remaining blocks */
            if (procnum == ((origproc + nprocs - 1) % nprocs))
            {
	    /* update on this processor in previous cycle */
/*
if (procnum == 0) {
  printf("\tupdate on this processor in previous cycle --- nlbrowUsend=%d - rowsfin=%d - j=%d - lbrowUloc=%d \n",
        nlbrowUsend,rowsfin,j,lbrowUloc);
  fflush(stdout); }
*/
               SparsekijLUUpdate((rowsfin + 1), rowsthisproc, &(lbrowU[lbrowUloc]), nlbrowUsend);
            }                   /* end if update in previous cycle */
            else if (procnum == origproc)       /* update on this processor
                                                 * in this cycle */
            {
               if ((rowsfin + jb + 1) <= rowsthisproc)
               {
/*
if (procnum == 0) {
  printf("\t update on this processor in this cycle --- nlbrowUrcv=%d - rowsfin=%d - j=%d - jb=%d \n",
        nlbrowUrcv,rowsfin,j,jb);
  fflush(stdout); }
*/
                  SparsekijLUUpdate((rowsfin + jb + 1), rowsthisproc, lbrowUrcv, nlbrowUrcv);
               }                /* end if */
            }                   /* end if update this cycle */
            else
            {
               /* otherwise */
/*
if (procnum == 0) {
  printf("\t otherwise --- nlbrowUrcv=%d - rowsfin=%d - j=%d - jb=%d \n", nlbrowUrcv,rowsfin,j,jb);
  fflush(stdout); }
*/
               SparsekijLUUpdate((rowsfin + 1), rowsthisproc, lbrowUrcv,
				 nlbrowUrcv);
            }                   /* end if otherwise */
         }                      /* end if (rowsthisproc-rowsfin) > 0 */
      }                         /* end for j */
   }                            /* end if parallel code */
}                               /* end BlockSparseFac() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparsekijLUSeq()
{
   int             i, j, k;
   int             locLik, locLij, endLij;
   int             locUkj, locUij, endUij;
   complex         c1;

   for (k = 1; k <= nlb; k++)   /* along the diagonal */
   {
      for (locUkj = (lbdiag[k - 1].PtrUix + 1); locUkj <= lbdiag[k].PtrUix; locUkj++)
      {
         /* scale */
         /* lbrowU[locUkj].Value /= lbdiag[k].Value; */
         division(lbrowU[locUkj].Value, lbdiag[k].Value, lbrowU[locUkj].Value);
      }                         /* end for locUkj */

/*      locLik = lbdiag[k].PtrLxj; */
      locLik = lbdiagcolpt[k].PtrLxj;
      while (locLik > 0)        /* column k */
      {
         i = lbrowL[locLik].i;
         locLij = (locLik + 1);
         endLij = lbdiag[i].PtrLix;
         locUij = (lbdiag[i - 1].PtrUix + 1);
         endUij = lbdiag[i].PtrUix;
         for (locUkj = (lbdiag[k - 1].PtrUix + 1); (locUkj <= lbdiag[k].PtrUix); locUkj++)      /* row k */
         {
            j = lbrowU[locUkj].j;

            if (i > j)          /* update the strictly lower triangular
                                 * matrix */
            {
               /* the value Lij must exist so search for it */
               while ((lbrowL[locLij].j < j) && (locLij < endLij))
                  locLij++;

               if (locLij > endLij)
               {
                  CMMD_error("error in SparsekijLUFac --- (locLij > endLij) - locLij=%d -  endLij=%d \n",
                     locLij, endLij);
               }

               /*
                * lbrowL[locLij].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbrowL[locLij].Value, c1, lbrowL[locLij].Value);

               nflopsl++;
            }                   /* end update the strictly lower triangular
                                 * matrix */

            else if (i == j)    /* update the diagonal */
            {

               /*
                * lbdiag[i].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbdiag[i].Value, c1, lbdiag[i].Value);

               nflopsd++;
            }                   /* end update the diagonal */
            else
            {
               /* if (i < j) - update the strictly upper triangular matrix */
               /* the value Uij must exist so search for it */
               while ((lbrowU[locUij].j < j) && (locUij < endUij))
                  locUij++;

               if (locUij > endUij)
               {
                  CMMD_error("error in SparsekijLUFac --- (locUij > endUij) - locUij=%d -  endUij=%d \n",
                     locUij, endUij);
               }

               /*
                * lbrowU[locUij].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbrowU[locUij].Value, c1, lbrowU[locUij].Value);

               nflopsu++;
            }                   /* end update the strictly upper triangular
                                 * matrix */
         }                      /* end for locUkj */
         locLik = lbcolL[locLik].PtrNext;
      }                         /* end while locLik > 0 */
   }                            /* end for k */
}                               /* end SparsekijLUSeq() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparsekijLUPar(kstart, jb)
int             kstart, jb;

{
   int             i, j, k, pi, pj;
   int             locLik, locLij, endLij;
   int             locUkj, locUij, endUij;
   complex         c1;

/*
printf("Enter SparsekijLUPar() --- kstart=%d - jb=%d - nflopsl=%d \n",kstart,jb,nflopsl);
fflush(stdout);
*/
   for (k = kstart; k <= (kstart + jb - 1); k++)        /* along the diagonal */
   {
      for (locUkj = (lbdiag[k - 1].PtrUix + 1); locUkj <= lbdiag[k].PtrUix; locUkj++)
      {
         /* scale */
         /* lbrowU[locUkj].Value /= lbdiag[k].Value; */
         division(lbrowU[locUkj].Value, lbdiag[k].Value, lbrowU[locUkj].Value);
      }                         /* end for locUkj */

      locLik = lbdiag[k].PtrLxj;
      while ((locLik > 0) && (lbrowL[locLik].i <= lbdiag[kstart + jb - 1].k))
      {
         /* column k */
         i = lbrowL[locLik].i;
         if ((((i - 1) / nb) % nprocs) != procnum)
         {
            CMMD_error("%d - error in SparsekijLUPar - value is not on the processor --- k=%d - locUkj=%d - j=%d \n",
               procnum, k, locUkj, j);
         }                      /* end if column j is on this processor */
         pi = ((nb * ((i - 1) / (nb * nprocs))) + ((i - 1) % nb) + 1);
         locLij = (locLik + 1);
         endLij = lbdiag[pi].PtrLix;
         locUij = (lbdiag[pi - 1].PtrUix + 1);
         endUij = lbdiag[pi].PtrUix;
         for (locUkj = (lbdiag[k - 1].PtrUix + 1); (locUkj <= lbdiag[k].PtrUix); locUkj++)      /* row k */
         {
            j = lbrowU[locUkj].j;
/*
printf("FAC --- i=%d - j=%d - k=%d - pi=%d - locLik=%d - locUkj=%d - locLij=%d - endLij=%d - locUij=%d - endUij=%d \n",
           i,j,k,pi,locLik,locUkj,locLij,endLij,locUij,endUij);
fflush(stdout);
*/
            if (i > j)          /* update the strictly lower triangular
                                 * matrix */
            {
               /* the value Lij must exist so search for it */
               while ((lbrowL[locLij].j < j) && (locLij < endLij))
                  locLij++;

               if (locLij > endLij)
               {
                  CMMD_error("%d - error in SparsekijLUFac --- (locLij > endLij) - locLij=%d -  endLij=%d \n",
                     procnum, locLij, endLij);
               }

               /*
                * lbrowL[locLij].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbrowL[locLij].Value, c1, lbrowL[locLij].Value);

               nflopsl++;
            }                   /* end update the strictly lower triangular
                                 * matrix */

            else if (i == j)    /* update the diagonal */
            {

               /*
                * lbdiag[pi].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbdiag[pi].Value, c1, lbdiag[pi].Value);

               nflopsd++;
            }                   /* end update the diagonal */

            else
            {
               /* if (i < j) - update the strictly upper triangular matrix */
               /* the value Uij must exist so search for it */
               while ((lbrowU[locUij].j < j) && (locUij < endUij))
                  locUij++;

               if (locUij > endUij)
               {
                  CMMD_error("%d - error in SparsekijLUFac --- (locUij > endUij) - locUij=%d -  endUij=%d \n",
                     procnum, locUij, endUij);
               }

               /*
                * lbrowU[locUij].Value -= lbrowL[locLik].Value *
                * lbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, lbrowU[locUkj].Value, c1);
               subtract(lbrowU[locUij].Value, c1, lbrowU[locUij].Value);

               nflopsu++;
            }                   /* end update the strictly upper triangular
                                 * matrix */
         }                      /* end for locUkj */
         locLik = lbcolL[locLik].PtrNext;
      }                         /* end while locLik > 0 */
   }                            /* end for k */
/*
printf("Exit SparsekijLUPar() --- kstart=%d - jb=%d \n",kstart,jb);
fflush(stdout);
*/
}                               /* end SparsekijLUPar() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparsekijLUUpdate(istart, iend, llbrowU, stvl)
int             istart, iend;

LBRow          *llbrowU;

int             stvl;

{
   int             i, j, k, m, pi;
   int             locLik, locLij, endLij;
   int             locUkj, locUkjRow, locUij, endUij;
   complex         c1;

/*
if (procnum != nprocs) {
  printf("Enter SparsekijLUUpdate() --- istart=%d - iend=%d - stvl=%d - nflopsl=%d \n",
                   istart,iend,stvl,nflopsl);
  fflush(stdout); }
*/
   locUkjRow = 0;
   for (m = 1; m <= nb; m++)
   {
      k = llbrowU[locUkjRow].i; /* 1 <= k <= nlb */
      locLik = lbdiagcolpt[k].PtrLxj;
      i = lbrowL[locLik].i;
      pi = ((nb * ((i - 1) / (nb * nprocs))) + ((i - 1) % nb) + 1);
/*
printf("\tm=%d - k=%d - locLik=%d - i=%d - pi=%d - locUkjRow=%d - istart=%d - iend=%d \n",
            m,k,locLik,i,pi,locUkjRow,istart,iend);
fflush(stdout);
*/
      while ((locLik > 0) && (pi < istart))     /* advance to istart */
      {
         locLik = lbcolL[locLik].PtrNext;
         i = lbrowL[locLik].i;
         pi = ((nb * ((i - 1) / (nb * nprocs))) + ((i - 1) % nb) + 1);
      }                         /* end while */
      while ((locLik > 0) && (pi <= iend))      /* calculate all rows <= iend */
      {
         if ((((i - 1) / nb) % nprocs) != procnum)
         {
            CMMD_error("%d - error in SparsekijLUUpdate - row is not on this processor --- i=%d - locLik=%d - k=%d \n",
               procnum, i, locLik, k);
         }                      /* end if row i is on this processor */
         locUkj = locUkjRow;
         locLij = (locLik + 1);
         endLij = lbdiag[pi].PtrLix;
         locUij = (lbdiag[pi - 1].PtrUix + 1);
         endUij = lbdiag[pi].PtrUix;
         while ((llbrowU[locUkj].i == k) && (locUkj <= stvl))
         {
            j = llbrowU[locUkj].j;
/*
printf("Update --- i=%d - j=%d - k=%d - pi=%d - locLik=%d - locUkj=%d - locLij=%d - endLij=%d - locUij=%d - endUij=%d \n",
           i,j,k,pi,locLik,locUkj,locLij,endLij,locUij,endUij);
fflush(stdout);
*/
            if (i > j)          /* update the strictly lower triangular
                                 * matrix */
            {
               /* the value Lij must exist so search for it */
               while ((lbrowL[locLij].j < j) && (locLij < endLij))
                  locLij++;

               if (locLij > endLij)
               {
                  CMMD_error("%d - error in SparsekijLUUpdate --- (locLij > endLij) - locLij=%d -  endLij=%d \n",
                     procnum, locLij, endLij);
               }

               /*
                * lbrowL[locLij].Value -= lbrowL[locLik].Value *
                * llbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, llbrowU[locUkj].Value, c1);
               subtract(lbrowL[locLij].Value, c1, lbrowL[locLij].Value);

               nflopsl++;
            }                   /* end update the strictly lower triangular
                                 * matrix */

            else if (i == j)    /* update the diagonal */
            {

               /*
                * lbdiag[pi].Value -= lbrowL[locLik].Value *
                * llbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, llbrowU[locUkj].Value, c1);
               subtract(lbdiag[pi].Value, c1, lbdiag[pi].Value);

               nflopsd++;
            }                   /* end update the diagonal */

            else
            {
               /* if (i < j) - update the strictly upper triangular matrix */
               /* the value Uij must exist so search for it */
               while ((lbrowU[locUij].j < j) && (locUij < endUij))
                  locUij++;

               if (locUij > endUij)
               {
                  CMMD_error("%d - error in SparsekijLUUpdate --- (locUij > endUij) - locUij=%d -  endUij=%d \n",
                     procnum, locUij, endUij);
               }

               /*
                * lbrowU[locUij].Value -= lbrowL[locLik].Value *
                * llbrowU[locUkj].Value;
                */
               multiply(lbrowL[locLik].Value, llbrowU[locUkj].Value, c1);
               subtract(lbrowU[locUij].Value, c1, lbrowU[locUij].Value);

               nflopsu++;
            }                   /* end update the strictly upper triangular
                                 * matrix */
            locUkj++;
         }                      /* end while (llbrowU(locUkj).i == k) &&
                                 * (locUkj <= stvl) */
         locLik = lbcolL[locLik].PtrNext;
         i = lbrowL[locLik].i;  /* if( locLik == 0) lbrowL[locLik].i = 0 */
         pi = ((nb * ((i - 1) / (nb * nprocs))) + ((i - 1) % nb) + 1);
      }                         /* end while (locLik > 0) && (pi <= iend) */

      while ((llbrowU[locUkjRow].i == k) && (locUkjRow < stvl))
         locUkjRow++;

   }                            /* end for m */
/*
printf("Exit SparsekijLUUpdate() --- istart=%d - iend=%d - stvl=%d \n",
                   istart,iend,stvl);
fflush(stdout);
*/
}                               /* end SparsekijLUUpdate() */

static int      send_count = 0;
static int      recv_count = 0;

/***********************************************************************
                             ACTIVESENDSLB
 ***********************************************************************/
void            activesendSLB(am, dir, bufsize)
LBRow          *am;             /* (input ) double array of data  */

int             bufsize;        /* buffer size to be sent */

char            dir;            /* direction to send data */

{

   int             destproc;

#ifdef SP1
   int             msgid;
   int             nbytes;

#endif

   if ((dir == 'N') || (dir == 'n'))
      destproc = ((procnum + 1) % nprocs);
   else if ((dir == 'P') || (dir == 'p'))
      destproc = ((procnum + nprocs - 1) % nprocs);

/* send the buffer to the next processor in the correct direction */

#if 0
#define DEBUG_activesendSLB
#endif

#ifdef DEBUG_activesendSLB
   printf("%d activesendSLB %d dest %d dir %c bufsize %d\n",
      CMMD_self_address(), ++send_count, destproc, dir, bufsize);
   fflush(stdout);
#undef DEBUG_activesendSLB
#endif

#if CMAML_RPC
   CMAML_disable_interrupts();

   CMAML_poll_while_lt(&rpc_ack, true); /* wait for receiver's signal */
   rpc_ack = false;

   if (destproc != dest_chk)    /* check the signal */
      CMMD_error("\n\nERROR - activesendSLB: Failed to receive request signal!\n");

   CMAML_scopy(destproc,
      rport,
      0,
      0,
      am,
      bufsize,
      0,
      0);

   CMAML_enable_interrupts();
#else

   CMMD_send(destproc, 1000, am, bufsize);
#endif

/*
printf("\tIn activesendSLB() send done --- procnum=%d - destproc=%d - dir=%c \n",
        procnum,destproc,dir);
fflush(stdout);
*/
}                               /* end ActiveSendSLB() */

/***********************************************************************
                             ACTIVERECVSLB
 ***********************************************************************/

void            activerecvSLB(am, dir, bufsize, origproc)
LBRow          *am;             /* (output)  array of data  */

int             bufsize, origproc;

char            dir;            /* direction to send data */

{
   int             srcproc, destproc;

#ifdef SP1
   int             type = 1000;
   int             msgid;
   int             nbytes;
#endif

   int             src_bufsize;

   if ((dir == 'N') || (dir == 'n'))
   {
      srcproc = ((procnum + nprocs - 1) % nprocs);
      destproc = ((procnum + 1) % nprocs);
   }
   else if ((dir == 'P') || (dir == 'p'))
   {
      srcproc = ((procnum + 1) % nprocs);
      destproc = ((procnum + nprocs - 1) % nprocs);
   }

#if 0
#define DEBUG_activerecvSLB
#endif

#ifdef DEBUG_activerecvSLB
   printf("%d activerecvSLB %d src %d dest %d dir %c bufsize %d\n",
      CMMD_self_address(), ++recv_count, srcproc, destproc, dir, bufsize);
   if (origproc != destproc)
      printf("%d activesendSLB %d dest %d dir %c bufsize %d\n",
         CMMD_self_address(), ++send_count, destproc, dir, bufsize);
   fflush(stdout);
#undef DEBUG_activerecvSLB
#endif

#if CMAML_RPC
   CMAML_disable_interrupts();

/* prepare rport for CMAML_scopy */
   CMAML_set_rport_addr(rport, am);
   CMAML_set_rport_byte_counter(rport, bufsize);
   CMAML_set_rport_total_nbytes(rport, bufsize);

/* send readiness signal to the sender */
   CMAML_rpc(srcproc, rpc_handler, procnum);

/* wait for receiving block data from the sender */
   CMAML_poll_while_lt(&scopy_ack, true);
   scopy_ack = false;

   CMAML_enable_interrupts();
/*
printf("\tIn activerecvSLB() receive done --- procnum=%d - srcproc=%d - destproc=%d - origproc=%d - dir=%c - bufsize=%d\n",
        procnum,srcproc,destproc,origproc,dir,bufsize);
fflush(stdout);
*/
/* if necessary, continue broadcast  - CMMD_bc has not been
   used because it is blocking */

   if (origproc != destproc)
   {
/*
printf("\tcontinue broadcast --- procnum=%d - srcproc=%d - destproc=%d - origproc=%d - dir=%c - bufsize=%d\n",
        procnum,srcproc,destproc,origproc,dir,bufsize);
fflush(stdout);
*/
      CMAML_disable_interrupts();

      CMAML_poll_while_lt(&rpc_ack, true);      /* wait for receiver's signal */
      rpc_ack = false;

      if (destproc != dest_chk) /* check the signal */
         CMMD_error("\n\nERROR - activerecvSLB: Failed to receive request signal!\n");

      CMAML_scopy(destproc,
         rport,
         0,
         0,
         am,
         bufsize,
         0,
         0);

      CMAML_enable_interrupts();
/*
printf("\tcontinue broadcast done --- procnum=%d - srcproc=%d - destproc=%d - origproc=%d - dir=%c - bufsize=%d\n",
        procnum,srcproc,destproc,origproc,dir,bufsize);
fflush(stdout);
*/
   }                            /* end if */
#else

   CMMD_receive(srcproc, 1000, am, bufsize);
   if (origproc != destproc)
   {
      CMMD_send(destproc, 1000, am, bufsize);
   }
#endif

/*
printf("\t => broadcast done --- procnum=%d - srcproc=%d - destproc=%d - origproc=%d - dir=%c\n",
        procnum,srcproc,destproc,origproc,dir);
fflush(stdout);
*/
}                               /* end ActiveRecvSLB() */

/***********************************************************************
                        Handler Function
 ***********************************************************************/

void            rpc_handler(pn) /* CMAML_rpc */
int             pn;

{
   dest_chk = pn;
   rpc_ack = true;
}

void            scopy_handler() /* CMAML_scopy */
{
   scopy_ack = true;
}

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ForwardReduction()
{
   int             i, l, iproc, bufsize, size;

/* solve for Ly=b */
/* reduce the independent blocks */
   CMMD_sync_with_host();
   if (procnum < nprocs)
   {
/*
for (i=1;i<=nlb;i++) {
   printf("lb[%d]=%lf\t",i,lb[i]);
   if ( (i%5) == 0) printf("\n"); }
printf("\n");
*/
      nflops = 0;
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      for (l = 1; l <= nblocks; l++)
      {

         ForwardReduceBlocks(l);

      }                         /* end for l */

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_FRIB, &node_tim, sizeof(node_timer));
   }                            /* end if an active processor */

/* reduce this processors portion of the lower border */
   CMMD_sync_with_host();

#if CMAML_RPC
   CMAML_disable_interrupts();
#endif

   if (procnum < nprocs)
   {
      if (CMAML_RPC == YES)
      {
         CMMD_node_timer_clear(0);
         CMMD_node_timer_start(0);

#if CMAML_RPC
         ForwardReduceBorderCMAML();
#endif

         CMMD_node_timer_stop(0);
      }                         /* end if CMAML_RP == YES */
      else
      {
         nlbupdatefr = (int *) calloc((nprocs + 1), sizeof(int));
         if (nlbupdatefr == NULL)
         {
            CMMD_error("can't allocate memory for nlbupdatefr[] \n");
         }                      /* end if */
         for (iproc = 0; iproc < nprocs; iproc++)
            nlbupdatefr[iproc] = 0;

         if (nprocs == 1)
            size = 1;
         else
            size = (nlb + 1);
         lbupdatefrtmp = (LBUpdateFr *) calloc((nlborder * (nprocs + 1)), sizeof(LBUpdateFr));
         if (lbupdatefrtmp == NULL)
         {
            CMMD_error("can't allocate memory for lbupdatefrtmp \n");
         }                      /* end if */

         lbupdatefr = (LBUpdateFr **) calloc((nprocs + 1), sizeof(LBUpdateFr *));
         if (lbupdatefr == NULL)
         {
            CMMD_error("can't allocate memory for lbupdatefr[] \n");
         }                      /* end if */
         for (i = 0; i < (nprocs + 1); i++)
            lbupdatefr[i] = lbupdatefrtmp + i * nlborder;

         bufsize = (nlb + 1);
         lbupdatefrrcv = (LBUpdateFr *) calloc(bufsize, sizeof(LBUpdateFr));
         lbupdatefrrcv_size = bufsize * sizeof(LBUpdateFr);
         if (lbupdatefrrcv == NULL)
         {
            CMMD_error("can't allocate memory for lbupdatefrrcv \n");
         }                      /* end if */

         CMMD_node_timer_clear(0);
         CMMD_node_timer_start(0);

         ForwardReduceBorder();

         CMMD_node_timer_stop(0);
         free(nlbupdatefr);
         free(lbupdatefrtmp);
         free(lbupdatefr);
         free(lbupdatefrrcv);
      }                         /* end if CMAML_RPC == NO */

      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_FRLB, &node_tim, sizeof(node_timer));
   }                            /* end if an active processor */

#if CMAML_RPC
   CMAML_enable_interrupts();
#endif

/* reduce this processors portion of the last block */
   CMMD_sync_with_host();
   if (procnum < nprocs)
   {
/*     nflops = 0; */

      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      ForwardReduceLastBlock();

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_FR, &node_tim, sizeof(node_timer));
/*
printf("FR --- NUMBER OF FLOATING POINT OPS FOR THE LAST BLOCK ON THIS PROCESSOR --- nflops=%d \n",nflops);
fflush(stdout);
*/
      printf("FR --- NUMBER OF FLOATING POINT OPS FOR THIS PROCESSOR --- nflops=%d \n", nflops);
      fflush(stdout);

#if 0
#define DEBUG_ForwardReduction
#endif

#ifdef DEBUG_ForwardReduction
      for (i = 1; i <= n; i++)
      {
         printf("%d ForwardReduction y %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, y[i].real, y[i].imag);
      }
      for (i = 1; i <= nlb; i++)
      {
         printf("%d ForwardReduction ly %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, ly[i].real, ly[i].imag);
      }
      fflush(stdout);
#undef DEBUG_ForwardReduction
#endif
   }                            /* end if an active processor */
}                               /* end ForwardReduction() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ForwardReduceBlocks(l)
int             l;

{
   int             i, j;
   int             locLij;
   complex         c1;

/*
printf("ForwardReduceBlocks(%d) \n",l);
fflush(stdout);
*/
   for (i = (blocks[l - 1].DiagPtr + 1); i <= blocks[l].DiagPtr; i++)
   {
      j = i - blocks[0].DiagPtr;
      for (locLij = (diag[j - 1].PtrLix + 1); locLij <= diag[j].PtrLix; locLij++)
      {

         /*
          * b[j] -= blockrow[locLij].Value * y[ (blockrow[locLij].j -
          * blocks[0].DiagPtr) ];
          */
         multiply(blockrow[locLij].Value, y[(blockrow[locLij].j - blocks[0].DiagPtr)], c1);
         subtract(b[j], c1, b[j]);

         nflops++;
      }                         /* end while */
/* This is Crout factorization, so all Uii are all equal to 1.0 */
      /* y[j] = b[j] / diag[j].Value; */
      division(b[j], diag[j].Value, y[j]);
   }                            /* end for i */

}                               /* end ForwardReduceBlocks() */

#if CMAML_RPC
/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ForwardReduceBorderCMAML()
{
   int             i, j, k, l;
   int             iproc;
   int             locLBij, endLBix;
   int             LBPtr, EndLBPtr;
   unsigned int    iupd1, iupd2;
   complex         update, iupd, c1;

/* reduce the border - except for the last lower diagonal block */
   for (l = 1; l <= nblocks; l++)
   {
/* initialize the pointers into the data structures */
      LBPtr = (blocks[l - 1].LBPtr + 1);
      EndLBPtr = blocks[l].LBPtr;

/* some lower border rows may be all zeros - so only use
   rows with nonzero values */
/*
printf("some lower border rows may be all zeros - so only use\n");
fflush(stdout);
*/
      while (LBPtr <= EndLBPtr)
      {
         update.real = 0.0;
         update.imag = 0.0;
         locLBij = (lborder[(LBPtr - 1)].PtrLBix + 1);
         endLBix = lborder[LBPtr].PtrLBix;

         while (locLBij <= endLBix)
         {

            /*
             * update += borderrow[locLBij].Value * y[ (borderrow[locLBij].j
             * - blocks[0].DiagPtr) ];
             */
            multiply(borderrow[locLBij].Value, y[(borderrow[locLBij].j - blocks[0].DiagPtr)], c1);
            add(update, c1, update);
            locLBij++;
            nflops++;
         }                      /* end while (locLBij <= endLix) */
         i = lborder[LBPtr].i - lastblockstart + 1;
         iproc = (((i - 1) % (nprocs * nb)) / nb);
         if (iproc != procnum)
         {
            iupd1 = (int) update.real;
            iupd2 = (int) update.imag;
            CMAML_rpc(iproc, StoreHandlerFRBCMAML, i, iupd1, iupd2);
         }                      /* end if */
         else
         {
            /* lb[i] -= update; */
            subtract(lb[i], update, lb[i]);
            ipollFRBCMAML++;
         }                      /* end else */
         CMAML_poll();
         LBPtr++;
      }                         /* end while (LBPtr <= EndLBPtr) */
   }                            /* end for l */
/* check to see if data has been received */
   CMAML_poll_while_lt(&ipollULBCMAML, nupdatesFRB[procnum + 1]);
/*
printf("Exit ForwardReduceBorderCMAML() - ipollFRBCMAML=%d - nupdatesFRB[%d]=%d \n",
           ipollFRBCMAML,(procnum+1),nupdatesFRB[procnum+1]);
fflush(stdout);
*/
}                               /* end ForwardReduceBorderCMAML() */

#endif

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            StoreHandlerFRBCMAML(i, iupd1, iupd2)
int             i;

unsigned int    iupd1, iupd2;

{
   complex         iupd;

   iupd.real = (float) iupd1;
   iupd.imag = (float) iupd2;
   /* lb[i] -= * ((double *) (&iupd)); */
   subtract(lb[i], iupd, lb[i]);
   ipollFRBCMAML++;
}                               /* end StoreHandlerFRBCMAML() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ForwardReduceBorder()
{
   int             i, j, k, l, bufsize;
   int             iproc, nextproc, prevproc, UplbStatus;
   int             locLBij, endLBix;
   int             LBPtr, EndLBPtr;
   complex         update, c1;
   CMMD_mcb        UplbMCB;

/* reduce the border - except for the last lower diagonal block */
   for (l = 1; l <= nblocks; l++)
   {
/* some lower border rows may be all zeros - so only use
   rows with nonzero values */
/*
printf("some lower border rows may be all zeros - so only use\n");
fflush(stdout);
*/
      for (LBPtr = (blocks[l - 1].LBPtr + 1); LBPtr <= blocks[l].LBPtr; LBPtr++)
      {
         update.real = 0.0;
         update.imag = 0.0;
         locLBij = (lborder[(LBPtr - 1)].PtrLBix + 1);
         endLBix = lborder[LBPtr].PtrLBix;

         while (locLBij <= endLBix)
         {

            /*
             * update += borderrow[locLBij].Value * y[ (borderrow[locLBij].j
             * - blocks[0].DiagPtr) ];
             */
            multiply(borderrow[locLBij].Value, y[(borderrow[locLBij].j - blocks[0].DiagPtr)], c1);
            add(update, c1, update);
            locLBij++;
            nflops++;
         }                      /* end while (locLBij <= endLix) */
         i = lborder[LBPtr].i - lastblockstart + 1;
         iproc = (((i - 1) % (nprocs * nb)) / nb);
         if (iproc != procnum)
         {
            lbupdatefr[iproc][nlbupdatefr[iproc]].Value.real = update.real;
            lbupdatefr[iproc][nlbupdatefr[iproc]].Value.imag = update.imag;
            lbupdatefr[iproc][nlbupdatefr[iproc]].i = i;
            (nlbupdatefr[iproc])++;
         }                      /* end if */
         else
         {
            /* lb[i] -= update; */
            subtract(lb[i], update, lb[i]);
         }                      /* end else */
      }                         /* end for LBPtr) */
   }                            /* end for l */
   for (iproc = 0; iproc < nprocs; iproc++)
   {
      lbupdatefr[iproc][nlbupdatefr[iproc]].i = 0;
      lbupdatefr[iproc][nlbupdatefr[iproc]].Value.real = 0.0;
      lbupdatefr[iproc][nlbupdatefr[iproc]].Value.imag = 0.0;
   }                            /* end for iproc */

/* send the appropriate data to the other processors - each processor
   will send data asychronously to another processor equidistant away -
   until all processors have talked to all other processsors */

   nextproc = procnum;
   prevproc = procnum;
   for (iproc = 0; iproc < (nprocs - 1); iproc++)
   {
      nextproc = ((nextproc + 1) % nprocs);
      prevproc = ((prevproc + nprocs - 1) % nprocs);
      bufsize = ((nlbupdatefr[nextproc] + 1) * sizeof(LBUpdateFr));
/*
printf("\t\t\tiproc=%d - nextproc=%d - prevproc=%d - nlbupdatefr[%d]=%d - (bytes)=%d \n",
          iproc,nextproc,prevproc,nextproc,nlbupdatefr[nextproc],bufsize);
fflush(stdout);
*/
      CMMD_send_async(nextproc, (SEND_UPDATELBFR + procnum), lbupdatefr[nextproc], bufsize, SendHandlerCMMD, NULL);

      bufsize = lbupdatefrrcv_size;

      UplbMCB = CMMD_receive_async(prevproc, (RECEIVE_UPDATELBFR + prevproc), lbupdatefrrcv, bufsize, ReceiveHandlerCMMD, &UplbStatus);
/* check to see if data has been received */
      if (UplbStatus != (RECEIVE_UPDATELBFR + prevproc))
      {
         CMMD_msg_wait(UplbMCB);
      }                         /* end if */

      k = 0;
      while ((lbupdatefrrcv[k].i != 0) && (lbupdatefrrcv[k].Value.real != 0.0))
      {
         i = lbupdatefrrcv[k].i;
         /* lb[i] -= lbupdatefrrcv[k].Value; */
         subtract(lb[i], lbupdatefrrcv[k].Value, lb[i]);
         k++;
      }                         /* end while */
   }                            /* end else (receive data) */

}                               /* end ForwardReduceBorder() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            ForwardReduceLastBlock()
{
   int             i, j;

/*
printf("in ForwardReduceLastBlock() - 1 \n");
fflush(stdout);
*/

   BlockSparseFr();

/*
printf("in ForwardReduceLastBlock() - 2 \n");
fflush(stdout);
*/
}                               /* end ForwardReduceLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
/* block forward reduction for sparse LU factored matrices with
   no inherient parallelism

  This code was inspired by hlltbk.f authored by Manish Parashar of NPAC */

void            BlockSparseFr()
{
/*
   nlb          is the number of rows in the matrix  (global)
   nb           is the number of rows in a block     (global)
   nprocs       is the number of processors          (global)
   rowsthisproc is the number of rows this processor (global))
*/
   int             j, jb, jb1, colptr, bkrowptr;
   int             proc, bufsize;

   colptr = bkrowptr = proc = 0;
/*
if (procnum == 1) {
  for (j=1;j<=nlb;j++) {
    printf("\t\tj=%d - lb[%d]=(%f,%f) \n",j,j,lb[j].real,lb[j].imag);
    fflush(stdout); } }
*/
/* for one processor --- set nprocs = 1 */
   if (nprocs == 1)             /* use unblocked, sequential code on
                                 * processor 0 */
   {
/*
printf("for one processor --- nb=%d - nlb=%d - procnum=%d \n",nb,nlb,procnum);
fflush(stdout);
*/
      SparseFrSeq();
   }                            /* end if one processor */
   else
   {
      /* use blocked, parallel code on processors less than nprocs */
/*
printf("for multiple  processors --- nb=%d - nlb=%d - procnum=%d \n",nb,nlb,procnum);
fflush(stdout);
*/
      if (procnum == 0)         /* Reduce the first block and send the values */
      {
         SparseFrPar(0, nb);
         bkrowptr += nb;        /* block row finished - so update bkrowptr */
         bufsize = (nb * sizeof(complex));
/*
printf("\tsend the values --- start_loc=%d - nb=%d - bufsize=%d - colptr=%d - bkrowptr=%d - bksthisproc=%d \n",
        (colptr+1),nb,bufsize,colptr,bkrowptr,bksthisproc);
fflush(stdout);
*/
         activesendSLB(&(ly[colptr + 1]), 'N', bufsize);
/*
printf("\tfinished sending the values --- start_loc=%d - nb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr+1),nb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
      }                         /* end if procnum == 0 */

      for (j = 1; j < nbks; j++)
      {
         jb = nb;
/*
printf("----->>>>> j=%d - nlb=%d - nb=%d -- jb=%d - nbks=%d <<<<<----- \n",j,nlb,nb,jb,nbks);
fflush(stdout);
*/
         if (proc != procnum)   /* receive the x-vector */
         {
            bufsize = (jb * sizeof(complex));
/*
printf("\treceive the values --- start_loc=%d - jb=%d - bufsize=%d \n",(colptr+1),jb,bufsize);
fflush(stdout);
*/
            activerecvSLB(&(ly[colptr + 1]), 'N', bufsize, proc);
/*
printf("\tfinish receiving the values --- start_loc=%d - jb=%d - bufsize=%d \n",(colptr+1),jb,bufsize);
fflush(stdout);
*/
         }                      /* end if origproc != procnum */

         proc = ((proc + 1) % nprocs);
         if (bkrowptr <= ((bksthisproc - 1) * nb))
         {
            if (proc == procnum)
            {
               if (j == (nbks - 1))
                  jb1 = (nlb - ((nbks - 1) * nb));
               else
                  jb1 = nb;
/*
printf("\t----->>>>> j=%d - nlb=%d - nb=%d -- jb1=%d  - bksthisproc=%d <<<<<----- \n",j,nlb,nb,jb1,bksthisproc);
fflush(stdout);
*/
               /* this processor is the next to forward reduce a panel */
               /* so only update the b-vector for the first panel */

               SparseFrUpdate(colptr, bkrowptr, (bkrowptr + jb), jb);

               /*
                * then forward reduce the values of the y-vector and send it
                */
               SparseFrPar(bkrowptr, jb1);
               bkrowptr += nb;
               if (j <= (nbks - 2))     /* send all but last block - avoid a
                                         * null send */
               {
                  bufsize = (jb1 * sizeof(complex));
/*
printf("\tsend the values --- start_loc=%d - jb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr+nb+1),jb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
                  activesendSLB(&(ly[colptr + nb + 1]), 'N', bufsize);
               }                /* end if */
            }                   /* end if proc == procnum */
/* update the remaining blocks */
            SparseFrUpdate(colptr, bkrowptr, rowsthisproc, jb);
            colptr += nb;       /* row block finished - so update colptr */
         }                      /* end if bkrowptr <= rowsthisproc */
         else
         {
            colptr += jb;
         }                      /* end if bkrowptr > rowsthisproc */
      }                         /* end for j */
   }                            /* end if parallel code */
/*
if (procnum == 1) {
  for (j=1;j<=nlb;j++) {
    printf("\t\tj=%d - ly[%d]=(%f,%f) - lb[%d]=(%f,%f) \n",j,j,ly[j].real,ly[j].imag,j,lb[j].real,lb[j].imag);
    fflush(stdout); } }
*/
}                               /* end BlockSparseFr() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparseFrSeq()
{
   int             j;
   int             locLxj;
   complex         c1;

/*
printf("Enter SparseFrSeq() --- \n");
fflush(stdout);
*/
   for (j = 1; j <= nlb; j++)   /* for all columns */
   {
      /* ly[j] = lb[j] / lbdiag[j].Value; */
      division(lb[j], lbdiag[j].Value, ly[j]);
      locLxj = lbdiag[j].PtrLxj;
      while (locLxj > 0)
      {
         /* lb[lbrowL[locLxj].i] -= ly[j] * lbrowL[locLxj].Value; */
         multiply(ly[j], lbrowL[locLxj].Value, c1);
         subtract(lb[lbrowL[locLxj].i], c1, lb[lbrowL[locLxj].i]);
         nflops++;
         locLxj = lbcolL[locLxj].PtrNext;
      }                         /* end while locLxj > 0 */
   }                            /* end for j */
/*
printf("Exit SparseFrSeq() --- \n");
fflush(stdout);
*/
}                               /* end SparseFrSeq() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
/* bkrowptr is the index corresponding to the blocked rowss */
void            SparseFrPar(bkrowptr, jb)
int             bkrowptr, jb;

{
   int             j, pj;
   int             locLxj;
   complex         c1;

/*
printf("Enter SparseFrPar() --- bkrowptr=%d - jb=%d \n",bkrowptr,jb);
fflush(stdout);
*/
   for (pj = (bkrowptr + 1); pj <= (bkrowptr + jb); pj++)
   {
      j = lbdiag[pj].k;
      /* ly[j] = lb[j] / lbdiag[pj].Value; */
      division(lb[j], lbdiag[pj].Value, ly[j]);
      locLxj = lbdiag[pj].PtrLxj;
      while ((locLxj > 0) && (lbrowL[locLxj].i <= lbdiag[bkrowptr + jb].k))
      {
         /* lb[lbrowL[locLxj].i] -= ly[j] * lbrowL[locLxj].Value; */
         multiply(ly[j], lbrowL[locLxj].Value, c1);
         subtract(lb[lbrowL[locLxj].i], c1, lb[lbrowL[locLxj].i]);
         nflops++;
         locLxj = lbcolL[locLxj].PtrNext;
      }                         /* end while */
   }                            /* end for i */
/*
printf("Exit SparseFrPar() --- bkrowptr=%d - jb=%d - nflops=%d \n",bkrowptr,jb,nflops);
fflush(stdout);
*/
}                               /* end SparseFrPar() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparseFrUpdate(colptr, bkrowptrstart, bkrowptrend, jb)
int             colptr, bkrowptrstart, bkrowptrend, jb;

{
   int             j, rowptrstart, rowptrend;
   int             locLxj;
   complex         c1;

   rowptrstart = (bkrowptrstart * nprocs) + (procnum * nb);
   rowptrend = (bkrowptrend * nprocs) + (procnum * nb);
/*
  printf("Enter SparseFrUpdate() --- colptr=%d - bkrowptrstart=%d - bkrowptrend=%d - nflops=%d \n",
                        colptr,bkrowptrstart,bkrowptrend,nflops);
  fflush(stdout);
*/
   for (j = (colptr + 1); j <= (colptr + jb); j++)
   {
      locLxj = lbdiagcolpt[j].PtrLxj;

      while ((lbrowL[locLxj].i < rowptrstart) && (locLxj > 0))
         locLxj = lbcolL[locLxj].PtrNext;

      while ((locLxj > 0) && (lbrowL[locLxj].i <= rowptrend))
      {
         /* lb[lbrowL[locLxj].i] -= ly[j] * lbrowL[locLxj].Value; */
         multiply(ly[j], lbrowL[locLxj].Value, c1);
         subtract(lb[lbrowL[locLxj].i], c1, lb[lbrowL[locLxj].i]);
         nflops++;
         locLxj = lbcolL[locLxj].PtrNext;
      }                         /* end while */
   }                            /* end for j */
/*
  printf("Exit SparseFrUpdate() --- colptr=%d - bkrowptrstart=%d - bkrowptrend=%d - nflops=%d \n",
                        colptr,bkrowptrstart,bkrowptrend,nflops);
  fflush(stdout);
*/
}                               /* end SparseFrUpdate() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            BackwardSubstitution()
{
   int             i, l;

/* do the backward substitution for the last block */
   CMMD_sync_with_host();
   if (procnum < nprocs)
   {
      nflops = 0;
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      BackSubLastBlock();

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_BSLB, &node_tim, sizeof(node_timer));
/*
printf("BS --- NUMBER OF FLOATING POINT OPS FOR THE LAST BLOCK ON THIS PROCESSOR --- nflops=%d \n",nflops);
fflush(stdout);
*/
   }                            /* end if an active processor */

/* do the backward substitution for the remaining blocks */
   CMMD_sync_with_host();
   if (procnum < nprocs)
   {
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      for (l = 1; l <= nblocks; l++)
      {

         BackSubBlocks(l);

      }                         /* end for l */

      CMMD_node_timer_stop(0);
      node_tim.pn_id = CMMD_self_address();
      node_tim.elapsed = CMMD_node_timer_elapsed(0);
      node_tim.busy = CMMD_node_timer_busy(0);
      node_tim.idle = CMMD_node_timer_idle(0);
      CMMD_send(CMMD_host_node(), SEND_TIME_BSIB, &node_tim, sizeof(node_timer));

      printf("BS --- NUMBER OF FLOATING POINT OPS FOR THIS PROCESSOR --- nflops=%d \n", nflops);

#if 0
#define DEBUG_BackwardSubstitution
#endif

#ifdef DEBUG_BackwardSubstitution
      for (i = 1; i <= n; i++)
      {
         printf("%d BackwardSubstitution x %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, x[i].real, x[i].imag);
         fflush(stdout);
      }
      for (i = 1; i <= nlb; i++)
      {
         printf("%d BackwardSubstitution lx %d (%9.6f, %9.6f)\n",
            CMMD_self_address(), i, lx[i].real, lx[i].imag);
         fflush(stdout);
      }
#endif

#undef DEBUG_BackwardSubstitution
   }                            /* end if an active processor */

}                               /* BackwardSubstitution() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            BackSubLastBlock()
{
/*
printf("in BackSubLastBlock() - 1 \n");
fflush(stdout);
*/

   BlockSparseBs();

}                               /* BackSubLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
/* block backward substitution for sparse LU factored matrices with
   no inherient parallelism

  This code was inspired by hlltbk.f authored by Manish Parashar of NPAC */

void            BlockSparseBs()
{
   int             j, jb, colptr, bkrowptr;
   int             proc, bufsize;

   proc = (nbks - 1) % nprocs;
   jb = nlb - ((nlb / nb) * nb);
   if (jb == 0)
      jb = nb;
   colptr = (nlb - jb);
   bkrowptr = (rowsthisproc - nb);      /* set the value of bkrowptr for all
                                         * procs but the last proc */
/*
printf("proc=%d - nlb=%d - nb=%d - jb=%d - colptr=%d - bkrowptr=%d - nbks=%d - rowsthisproc=%d\n",
                       proc,nlb,nb,jb,colptr,bkrowptr,nbks,rowsthisproc);
fflush(stdout);
*/
/* for one processor --- set nprocs = 1 */
   if (nprocs == 1)             /* use unblocked, sequential code on
                                 * processor 0 */
   {
/*
printf("for one processor --- jb=%d - nb=%d - nlb=%d \n",jb,nb,nlb);
fflush(stdout);
*/
      SparseBsSeq();
   }                            /* end if one processor */
   else
   {
      /* use blocked, parallel code on processors less than nprocs */
/*
printf("for multiple  processors --- jb=%d - nb=%d - nlb=%d \n",jb,nb,nlb);
fflush(stdout);
*/
      if (procnum == proc)      /* Reduce the first block and send the values */
      {
         bkrowptr = (rowsthisproc - jb);        /* set the value of bkrowptr
                                                 * for the last proc */
         SparseBsPar(bkrowptr, jb);
         bkrowptr -= nb;        /* column block finished - so update bkrowptr */
         bufsize = (jb * sizeof(complex));
/*
printf("\tsend the values --- start_loc=%d - jb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr+1),jb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
         activesendSLB(&(lx[colptr + 1]), 'P', bufsize);
/*
printf("\tfinished sending the values --- start_loc=%d - jb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr+1),jb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
      }                         /* end if procnum == proc */

      for (j = 1; j <= nbks; j++)
      {
         if (j > 1)
            jb = nb;
         if (proc != procnum)   /* receive the x-vector */
         {
            bufsize = (jb * sizeof(complex));
/*
printf("\treceive the values --- start_loc=%d - jb=%d - bufsize=%d \n",(colptr+1),jb,bufsize);
fflush(stdout);
*/
            activerecvSLB(&(lx[colptr + 1]), 'P', bufsize, proc);
/*
printf("\tfinish receiving the values --- start_loc=%d - jb=%d - bufsize=%d \n",(colptr+1),jb,bufsize);
fflush(stdout);
*/
         }                      /* end if proc != procnum */
         proc = ((proc + nprocs - 1) % nprocs);

         if (bkrowptr >= 0)
         {
            if (proc == procnum)
            {

               /*
                * this processor is the next to backward substitute a panel
                */
               /* so only update the y-vector for first panel */

               SparseBsUpdate(colptr, bkrowptr, bkrowptr, nb);

               /* then substitute the values of the x-vector and send it */
               SparseBsPar(bkrowptr, nb);
               /* column block finished - so update bkrowptr */
               bkrowptr -= nb;
               bufsize = (nb * sizeof(complex));
/*
printf("\tsend the values --- start_loc=%d - jb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr-nb+1),jb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
               activesendSLB(&(lx[colptr - nb + 1]), 'P', bufsize);
/*
printf("\tfinished sending the values --- start_loc=%d - jb=%d - bufsize=%d - colptr=%d - bkrowptr=%d \n",
        (colptr-nb+1),jb,bufsize,colptr,bkrowptr);
fflush(stdout);
*/
            }                   /* end if proc == procnum */

            SparseBsUpdate(colptr, bkrowptr, 0, jb);

            colptr -= nb;       /* row block finished - so update colptr */
         }                      /* end if bkrowptr >= 0 */
         else
         {
            colptr -= nb;
         }                      /* end if bkrowptr < 0 */
      }                         /* end for j */
   }                            /* end if parallel code */
/*
if (procnum == 0) {
  for (j=1;j<=nlb;j++) {
    printf("\t\tj=%d - lx[%d]=(%f,%f) - ly[%d]=(%f,%f) - lb[%d]=(%f,%f) \n",
         j,j,lx[j].real,lx[j].imag,j,ly[j].real,ly[j].imag,j,lb[j].real,lb[j].imag);
    fflush(stdout); } }
*/
}                               /* end BlockSparseBs() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparseBsSeq()
{
   int             j;
   int             locUxj;
   complex         c1;

   for (j = nlb; j >= 1; j--)   /* for all columns */
   {
      lx[j] = ly[j];
      locUxj = lbdiag[j].PtrUxj;
      while (locUxj > 0)
      {
         /* ly[lbrowU[locUxj].j] -= lx[j] * lbrowU[locUxj].Value; */
         multiply(lx[j], lbrowU[locUxj].Value, c1);
         subtract(ly[lbrowU[locUxj].j], c1, ly[lbrowU[locUxj].j]);
         nflops++;
         locUxj = lbcolU[locUxj].PtrNext;
      }                         /* end while locUxj > 0 */
   }                            /* end for j */
}                               /* end SparseBsSeq() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
/* bkrowptr is the index corresponding to the blocked rowss */
void            SparseBsPar(bkrowptr, jb)
int             bkrowptr, jb;

{
   int             j, pj;
   int             locUxj;
   complex         c1;

/*
printf("Enter SparseBsPar() --- bkrowptr=%d - jb=%d - nflops=%d \n",bkrowptr,jb,nflops);
fflush(stdout);
*/
   for (pj = (bkrowptr + jb); pj >= (bkrowptr + 1); pj--)
   {
      j = lbdiag[pj].k;
      lx[j] = ly[j];
      locUxj = lbdiag[pj].PtrUxj;
      while ((locUxj > 0) && (lbrowU[locUxj].i >= lbdiag[bkrowptr + 1].k))
      {
         /* ly[lbrowU[locUxj].i] -= lx[j] * lbrowU[locUxj].Value; */
         multiply(lx[j], lbrowU[locUxj].Value, c1);
         subtract(ly[lbrowU[locUxj].i], c1, ly[lbrowU[locUxj].i]);
         nflops++;
         locUxj = lbcolU[locUxj].PtrNext;
      }                         /* end for locUxj */
   }                            /* end for i1 */
/*
printf("Exit SparseBsPar() --- bkrowptr=%d - jb=%d - nflops=%d \n",bkrowptr,jb,nflops);
fflush(stdout);
*/
}                               /* end SparseBsPar() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            SparseBsUpdate(colptr, bkrowptrstart, bkrowptrend, jb)
int             colptr, bkrowptrstart, bkrowptrend, jb;

{
   int             j, rowptrstart, rowptrend;
   int             locUxj;
   complex         c1;

   rowptrstart = (bkrowptrstart * nprocs) + (procnum * nb);
   rowptrend = (bkrowptrend * nprocs) + (procnum * nb);
/*
printf("Enter SparseBsUpdate() --- colptr=%d - bkrowptrstart=%d - bkrowptrend=%d - jb=%d - rowptrstart=%d - rowptrend=%d \n",
                   colptr,bkrowptrstart,bkrowptrend,jb,rowptrstart,rowptrend);
fflush(stdout);
*/
   for (j = (colptr + 1); j <= (colptr + jb); j++)
   {
      locUxj = lbdiagcolpt[j].PtrUxj;
/*
printf("\tcolptr=%d - j1=%d - bkrowptrend=%d - locUxj=%d - lbrowU[%d].j=%d \n",
                   colptr,j1,bkrowptrend,locUxj,locUxj,lbrowU[locUxj].j);
fflush(stdout);
*/
      while ((lbrowU[locUxj].i > (rowptrstart + nb)) && (locUxj > 0))
         locUxj = lbcolU[locUxj].PtrNext;

      while ((locUxj > 0) && (lbrowU[locUxj].i > rowptrend))
      {
         /* ly[lbrowU[locUxj].i] -= lx[j] * lbrowU[locUxj].Value; */
         multiply(lx[j], lbrowU[locUxj].Value, c1);
         subtract(ly[lbrowU[locUxj].i], c1, ly[lbrowU[locUxj].i]);
         nflops++;
         locUxj = lbcolU[locUxj].PtrNext;
      }                         /* end while */
   }                            /* end for i1 */
/*
printf("Exit SparseBsUpdate() --- colptr=%d - bkrowptrstart=%d - bkrowptrend=%d - jb=%d - rowptrstart=%d - rowptrend=%d \n",
                   colptr,bkrowptrstart,bkrowptrend,jb,rowptrstart,rowptrend);
fflush(stdout);
*/
}                               /* end SparseBsUpdate() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void            BackSubBlocks(l)
int             l;

{
   int             i, j;
   int             UBPtr, EndUBPtr;
   int             locUBij, endUBxj;
   int             locUij, endUxj;
   complex         c1;

/* update y[] using the values in the lth border section */

/* initialize the pointers into the data structures */
   UBPtr = blocks[l].UBPtr;
   EndUBPtr = (blocks[l - 1].UBPtr + 1);

/* some upper border rows may be all zeros - so only use
   rows with nonzero values */
   while (UBPtr >= EndUBPtr)
   {
      j = uborder[UBPtr].j - lastblockstart + 1;
      locUBij = (uborder[(UBPtr - 1)].PtrUBxj + 1);
      endUBxj = uborder[UBPtr].PtrUBxj;

      while (locUBij <= endUBxj)
      {

         /*
          * y[ (bordercol[locUBij].i - blocks[0].DiagPtr)] -=
          * bordercol[locUBij].Value * lx[j];
          */
         multiply(bordercol[locUBij].Value, lx[j], c1);
         subtract(y[(bordercol[locUBij].i - blocks[0].DiagPtr)], c1, y[(bordercol[locUBij].i - blocks[0].DiagPtr)]);
         locUBij++;
         nflops++;
      }                         /* end while (locUBij <= endUBxj) */
      UBPtr--;
   }                            /* end while (UBPtr >= EndUBPtr) */

/* back substitute for the lth block */

   for (i = blocks[l].DiagPtr; i >= (blocks[l - 1].DiagPtr + 1); i--)
   {
      j = i - blocks[0].DiagPtr;
/* This is Crout factorization, so the values of Ujj are all equal to 1.0 */
      x[j] = y[j];

      locUij = (diag[j - 1].PtrUxj + 1);
      endUxj = diag[j].PtrUxj;
      while (locUij <= endUxj)
      {

         /*
          * y[ (blockcol[locUij].i - blocks[0].DiagPtr) ] -=
          * blockcol[locUij].Value * x[j];
          */
         multiply(blockcol[locUij].Value, x[j], c1);
         subtract(y[(blockcol[locUij].i - blocks[0].DiagPtr)], c1, y[(blockcol[locUij].i - blocks[0].DiagPtr)]);
         locUij++;
         nflops++;
      }                         /* end while */
   }                            /* end for i */

}                               /* end BackSubBlocks() */
